import pytest

from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import RandomOverSampler
from copy import deepcopy

from edgaro.balancing.transformer import TransformerFromIMBLEARN, SMOTE
from edgaro.data.dataset import Dataset

from .resources.objects import *


@pytest.fixture(scope='module', params=[
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
def sampler(request):
    return request.param


@pytest.fixture(scope='module', params=[
    (name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
    (name_2, pd.concat([df_2 for _ in range(5)]), pd.concat([target_2 for _ in range(5)]))])
def dataset(request):
    return Dataset(request.param[0], request.param[1], request.param[2])


def test_transformer(sampler, dataset):
    try:
        transformer = TransformerFromIMBLEARN(sampler)
        transformer.fit(dataset)
        transformer.transform(dataset)
        str(transformer)
        repr(transformer)
    except (Exception,):
        assert False


def test_transformer_train_test_split(sampler, dataset):
    ds = deepcopy(dataset)
    ds.train_test_split()

    try:
        transformer = TransformerFromIMBLEARN(sampler)
        transformer.fit(ds)
        transformer.transform(ds)
        str(transformer)
        repr(transformer)
    except (Exception,):
        assert False


def test_sufix(sampler, dataset):
    transformer = TransformerFromIMBLEARN(sampler)
    transformer.fit(dataset)
    out = transformer.transform(dataset)

    assert out.name == dataset.name + '_transformed'


@pytest.mark.parametrize('sufix', [
    'sufix_1',
    'sufix_2',
    'example_sufix'
])
def test_sufix_2(sampler, dataset, sufix):
    transformer = TransformerFromIMBLEARN(sampler, sufix)
    transformer.fit(dataset)
    out = transformer.transform(dataset)

    assert out.name == dataset.name + sufix


@pytest.mark.parametrize('imblearn_sampler,expected_imbalance_ratio', [
    (RandomUnderSampler(sampling_strategy=1, random_state=42), 1),
    (RandomOverSampler(sampling_strategy=1, random_state=42), 1)
])
def test_imbalance_ratio(imblearn_sampler, expected_imbalance_ratio):
    ds = Dataset(name_1, df_1, target_1)
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    transformer.fit(ds)
    out = transformer.transform(ds)

    assert out.imbalance_ratio == expected_imbalance_ratio


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42)
])
def test_was_fitted(imblearn_sampler):
    ds = Dataset(name_1, df_1, target_1)
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    assert not transformer.was_fitted
    transformer.fit(ds)
    assert transformer.was_fitted


def test_output_verbose(sampler, dataset, capsys):
    transformer = TransformerFromIMBLEARN(sampler, verbose=True)
    transformer.fit(dataset)
    transformer.transform(dataset)

    captured = capsys.readouterr()
    assert f'Transformer {transformer.__repr__()} was fitted with {dataset.name}' in captured.out
    assert f'Transformer {transformer.__repr__()} transformed with {dataset.name}' in captured.out


def test_empty_data_fit(sampler, dataset):
    ds = deepcopy(dataset)
    ds.data = None

    transformer = TransformerFromIMBLEARN(sampler)
    with pytest.raises(Exception):
        transformer.fit(ds)


def test_empty_data_target(sampler, dataset):
    ds = deepcopy(dataset)
    ds.target = None

    transformer = TransformerFromIMBLEARN(sampler)
    with pytest.raises(Exception):
        transformer.fit(ds)


@pytest.mark.parametrize('ds', [
    Dataset(name_2, pd.concat([df_2 for _ in range(10)]), pd.concat([target_2 for _ in range(10)])),  # all continuous
    Dataset(name_1, pd.concat([df_1 for _ in range(10)]), pd.concat([target_1 for _ in range(10)])),  # mixed types
    Dataset(name_1, pd.concat([df_1_categorical for _ in range(10)]), pd.concat([target_1 for _ in range(10)])),  # all categorical
])
def test_smote(ds):
    try:
        transformer = SMOTE()
        transformer.fit(ds)
        transformer.transform(ds)
        str(transformer)
        repr(transformer)
    except (Exception,):
        assert False



from __future__ import annotations

from typing import List, Dict, Optional, Any, Union

from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.base.base_transformer_array import BaseTransformerArray
from edgaro.balancing.transformer import Transformer
from edgaro.base.utils import print_unbuffered


class TransformerArray(BaseTransformerArray):
    """
    Create a class to apply Transformer transformation with more than one set of parameters and/or
    to each of the Dataset objects in DatasetArray.

    Note: If you use NestedAutomaticTransformer (or children class) as a parameter to TransformerArray, it is advisable
    to pass parameter `set_suffixes=False` in TransformerArray object. Otherwise, the suffixes will be distorted.

    Parameters
    ----------
    base_transformer : Transformer
        The object defining the transformation procedure.
    parameters : list[list, Dict[str, Any]]], optional
        The list of parameters for base_transformer. If the object is used for a DatasetArray object,
        the parameter list should be nested. For details, see Examples section.
    keep_original_dataset : bool, default=False
        Keep the original Dataset after transformations or not.
    dataset_suffixes : list, str, default='_transformed'
        Suffixes to be set to a transformed objects.
    result_array_sufix : str, default='_transformed_array'
        Suffix of the main transformed DatasetArray object.
    allow_dataset_array_sufix_change : bool, default=True
        Allow changing passed value of `result_array_sufix` according to `dataset_suffixes`.
    verbose : bool, default=False
        Print messages during calculations.
    set_suffixes : str, default=True
        Information whether suffixes for sub-Transformers should be set.

    Examples
    ----------
    Example 1

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.balancing.transformer import RandomUnderSampler
    >>> from edgaro.balancing.transformer_array import TransformerArray
    >>> df = Dataset(name_1, df_1, target_1)
    >>> params = [{'sampling_strategy': 0.98}, {'sampling_strategy': 1}]
    >>> transformer = RandomUnderSampler()
    >>> array = TransformerArray(transformer, parameters=params)
    >>> array.fit(df)
    >>> array.transform(df)

    Example 2

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.balancing.transformer import RandomUnderSampler
    >>> from edgaro.balancing.transformer_array import TransformerArray
    >>> df = DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)])
    >>> params = [ [{'sampling_strategy': 0.98}, {'sampling_strategy': 1}] for _ in range(len(df)) ]
    >>> transformer = RandomUnderSampler()
    >>> array = TransformerArray(transformer, parameters=params)
    >>> array.fit(df)
    >>> array.transform(df)
    """
    def __init__(self, base_transformer: Transformer, parameters: Optional[List[Union[List, Dict[str, Any]]]] = None,
                 keep_original_dataset: bool = False, dataset_suffixes: Union[str, List[str]] = '_transformed',
                 result_array_sufix: str = '_transformed_array', allow_dataset_array_sufix_change: bool = True,
                 verbose: bool = False, set_suffixes: bool = True) -> None:
        super().__init__(base_transformer=base_transformer, parameters=parameters, transformer_sufix=result_array_sufix)
        self.__dataset_suffixes = None

        self.set_suffixes = set_suffixes

        self.set_dataset_suffixes(dataset_suffixes)
        self.keep_original_dataset = keep_original_dataset
        self.allow_dataset_array_sufix_change = allow_dataset_array_sufix_change
        self.verbose = verbose

    def set_dataset_suffixes(self, name_sufix: Union[str, List[str]]) -> None:
        """
        Set suffixes to be set to transformed Dataset.

        Parameters
        ----------
        name_sufix : str, list
            Suffixes to be set to a transformed Dataset.
        """
        if self.set_suffixes:
            params = self.get_params()
            length_params = len(params) if params is not None else 0
            set_names = True if len(self.transformers) > 0 else False
            transformers = self.transformers

            if params is None:
                if isinstance(name_sufix, str):
                    self.__dataset_suffixes = [name_sufix]
                    if set_names:
                        transformers[0].set_dataset_suffixes(name_sufix)

                elif isinstance(name_sufix, list) and len(name_sufix) == 1 and isinstance(name_sufix[0], str):
                    self.__dataset_suffixes = name_sufix
                    if set_names:
                        transformers[0].set_dataset_suffixes(name_sufix)

                elif isinstance(name_sufix, list) and (length_params == 0 or len(name_sufix) == length_params):
                    self.__dataset_suffixes = name_sufix
                    if set_names:
                        for i in range(len(transformers)):
                            transformers[i].set_dataset_suffixes(name_sufix[i])

                else:
                    raise Exception('Wrong sufix names!')

            else:
                if isinstance(name_sufix, list) and len(name_sufix) == 1 and isinstance(name_sufix[0], str):
                    name_sufix = name_sufix[0]

                if isinstance(name_sufix, str):
                    if length_params == 1:
                        self.__dataset_suffixes = [name_sufix]
                        if set_names:
                            transformers[0].set_dataset_suffixes(name_sufix)

                    else:
                        self.__dataset_suffixes = [name_sufix + '_' + str(i) for i in range(length_params)]
                        if set_names:
                            for i in range(len(transformers)):
                                transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])

                elif isinstance(name_sufix, list) and len(name_sufix) == length_params:
                    self.__dataset_suffixes = name_sufix
                    if set_names:
                        for i in range(len(transformers)):
                            transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])

                else:
                    raise Exception('Parameter dataset_suffixes has invalid length!')

    def get_dataset_suffixes(self) -> Optional[Union[str, List[str]]]:
        """
        Get suffixes for transformed Dataset.

        Returns
        -------
        str, list
            Suffixes for a transformed Dataset.
        """
        return self.__dataset_suffixes

    def set_params(self, **params) -> None:
        """
        Set params for Transformer.

        Parameters
        ----------
        params : dict
            The parameters to be set.
        """
        super().set_params(**params)
        if self.__dataset_suffixes is not None and len(self.__dataset_suffixes) == 1:
            self.set_dataset_suffixes(self.__dataset_suffixes[0])

    def fit(self, dataset: Union[Dataset, DatasetArray]) -> None:
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            The object to fit Transformer on.
        """
        super().fit(dataset)
        self.__fix_classes()

        if self.set_suffixes:
            # Setting suffixes
            params = self.get_params()

            # Dataset case
            if isinstance(dataset, Dataset):
                if params is None:
                    self.transformers[0].set_dataset_suffixes(self.__dataset_suffixes[0])
                elif len(self.__dataset_suffixes) == len(self.transformers):
                    for i in range(len(self.transformers)):
                        self.transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])
                else:
                    raise Exception('Wrong length of dataset_suffixes!')
            # DatasetArray case
            else:
                if params is None:
                    for i in range(len(self.transformers)):
                        if len(self.__dataset_suffixes) == 1:
                            self.transformers[i].set_dataset_suffixes(self.__dataset_suffixes[0])
                        else:
                            self.transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])
                else:
                    for i in range(len(dataset)):
                        for j in range(len(self.transformers[i])):
                            if len(self.__dataset_suffixes) == 1:
                                self.transformers[i][j].set_dataset_suffixes(self.__dataset_suffixes[0] + '_' + str(j))
                            elif len(self.__dataset_suffixes[i]) == len(self.transformers[i]):
                                self.transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])
                            elif len(self.__dataset_suffixes) == len(self.transformers[i]):
                                self.transformers[i].set_dataset_suffixes(self.__dataset_suffixes[i])
                            else:
                                raise Exception('Wrong length of dataset_suffixes!')

            if isinstance(self.__dataset_suffixes, list) and len(self.__dataset_suffixes) == 1 and isinstance(
                    self.__dataset_suffixes[0], str):
                if self.allow_dataset_array_sufix_change:
                    self.transformer_sufix = self.__dataset_suffixes[0]

        if self.verbose:
            print_unbuffered(f'TransformerArray {self.__repr__()} was fitted with {dataset.name}')

    def transform(self, dataset: Union[Dataset, DatasetArray]) -> Union[Dataset, DatasetArray]:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            The object to be transformed.

        Returns
        -------
        Dataset, DatasetArray
            The transformed object.
        """
        out = super().transform(dataset=dataset)
        if self.keep_original_dataset:
            out.append(dataset)

        if self.verbose:
            print_unbuffered(f'TransformerArray {self.__repr__()} transformed with {dataset.name}')

        return out

    def __fix_classes(self) -> None:
        for i in range(len(self.transformers)):
            self.transformers[i] = self.__base_transformer_array_to_balancing_transformer_array(
                self.transformers[i])

    def __base_transformer_array_to_balancing_transformer_array(self, base: Union[
            Transformer, TransformerArray, BaseTransformerArray]) -> Union[Transformer, TransformerArray]:
        if isinstance(base, Transformer):
            return base
        elif not isinstance(base, TransformerArray) and isinstance(base, BaseTransformerArray):
            out = TransformerArray(base_transformer=self.base_transformer)
            out.__class__ = self.__class__
            for key, val in base.__dict__.items():
                out.__dict__[key] = val

            return out
        else:
            return base

    @property
    def transformers(self) -> List[Union[Transformer, TransformerArray, List]]:
        """
        All the Transformer objects used by this object.

        Returns
        -------
        list[Transformer, TransformerArray, list]
        """
        return super().transformers

    @transformers.setter
    def transformers(self, val: List[Union[Transformer, TransformerArray, List]]) -> None:
        super().transformers = val

    @property
    def base_transformer(self) -> Transformer:
        """
        Base transformers for creation of this object.

        Returns
        -------
        Transformer
        """
        out = super().base_transformer
        if isinstance(out, Transformer):
            return out
        else:
            raise Exception('Wrong base_transformer attribute')

    @base_transformer.setter
    def base_transformer(self, val: Transformer) -> None:
        BaseTransformerArray.base_transformer.fset(self, val)

    def __str__(self) -> str:
        return f"TransformerArray{(' ' + self.__class__.__name__) if self.__class__.__name__ != 'TransformerArray' else ''} with {len(self.transformers)} transformers"

    def __repr__(self) -> str:
        return f"<TransformerArray{(' ' + self.__class__.__name__) if self.__class__.__name__ != 'TransformerArray' else ''} with {len(self.transformers)} transformers>"



from __future__ import annotations

import multiprocessing

from typing import Union, Optional, Literal

from edgaro.model.model import Model
from edgaro.model.model_array import ModelArray
from edgaro.explain.explainer import Explainer
from edgaro.explain.explainer_result_array import ExplanationArray, ModelProfileExplanationArray, \
    ModelPartsExplanationArray
from edgaro.base.utils import print_unbuffered


class ExplainerArray:
    """
    Create a class to calculate PDP [1]_, ALE [2]_ curves or Variable Importance for Model and ModelArray objects.

    Parameters
    ----------
    models : Model, ModelArray
        A Model/ModelArray object to calculate the curves on.
    N : int, optional, default=None
        Number of observations that will be sampled from the test Dataset before the calculation of profiles
        (PDP/ALE curves). None means all data.
    explanation_type : {'PDP', 'ALE', 'VI'}, default='PDP'
        An explanation type to be calculated (`PDP` - Partial Dependence Profile, `ALE` - Accumulated Local Effects,
        `VI` - Variable Importance)
    verbose : bool, default=False
        Print messages during calculations.
    processes : int, default=1
        Number of processes for the calculation of explanations.
        If -1, it is replaced with the number of available CPU cores.
    random_state : int, optional, default=None
        Random state seed.
    B : int, optional, default=10
        Number of permutation rounds to perform on each variable - applicable only if explanation_type='VI'.

    Attributes
    ----------
    models : Model, ModelArray
        A Model/ModelArray object to calculate the curves for.
    name: str
        A name of the ExplainerArray, by default it is a Model/ModelArray name.
    sub_calculators : list[Explainer, ExplainerArray], optional
        A list of calculators for nested Datasets/DatasetArrays.
    N : int, optional
        Number of observations that will be sampled from the test Dataset before the calculation of profiles
        (PDP/ALE curves). None means all data.
    explanation_type : {'PDP', 'ALE', 'VI'}, default='PDP'
        An explanation type to be calculated.
    verbose : bool
        Print messages during calculations.
    processes : int
        Number of processes for the calculation of explanations.
        If -1, it is replaced with the number of available CPU cores.
    random_state : int, optional
        Random state seed
    B : int, optional
        Number of permutation rounds to perform on each variable - applicable only if explanation_type='VI'.

    """

    def __init__(self, models: Union[Model, ModelArray], N: Optional[int] = None,
                 explanation_type: Literal['PDP', 'ALE', 'VI'] = 'PDP', verbose: bool = False, processes: int = 1,
                 random_state: Optional[int] = None, B: Optional[int] = 10) -> None:
        self.models = models
        self.sub_calculators = None
        self.name = models.name
        self.N = N
        self.explanation_type = explanation_type
        self.verbose = verbose

        self.processes = processes
        self.random_state = random_state
        self.B = B

        if self.processes == -1:
            self.processes = multiprocessing.cpu_count()

    def fit(self) -> None:
        """
        Fit the ExplainerArray object and create an explainer attribute.
        """

        def create_sub_calculator(model: Union[Model, ModelArray]):
            if isinstance(model, Model):
                calc = Explainer(model=model, N=self.N, explanation_type=self.explanation_type, verbose=self.verbose,
                                 processes=self.processes, random_state=self.random_state, B=self.B)
            else:
                calc = ExplainerArray(models=model, N=self.N, explanation_type=self.explanation_type,
                                      verbose=self.verbose, processes=self.processes, random_state=self.random_state,
                                      B=self.B)

            calc.fit()
            return calc

        self.sub_calculators = [create_sub_calculator(m) for m in self.models.get_models()]

        if self.verbose:
            print_unbuffered(f'dalex explainers inside {self.__repr__()} were created')

    def transform(self, variables=None) -> ExplanationArray:
        """
        Calculate the explanation.

        Parameters
        ----------
        variables : list[str], optional
            List of variables for which the explanation should be calculated.

        Returns
        -------
        ExplanationArray
        """
        if self.verbose:
            print_unbuffered(f'{self.explanation_type}s are being calculated in {self.__repr__()}')

        if self.sub_calculators is None:
            raise Exception('Explainer was not fitted!')

        if variables is None:
            res = [calc.transform() for calc in self.sub_calculators]
        else:
            def transform_given_variables(calc):
                if isinstance(calc, Explainer):
                    var = list(set(variables).intersection(calc.model.get_train_dataset().data.columns))
                    return calc.transform(variables=var)
                else:
                    return calc.transform(variables=variables)

            res = [transform_given_variables(calc) for calc in self.sub_calculators]

        if len(res) == 1:
            out = res[0]
        else:
            cls = ModelProfileExplanationArray if self.explanation_type == 'PDP' or \
                                                  self.explanation_type == 'ALE' else ModelPartsExplanationArray
            out = cls(
                results=res,
                name=self.name,
                explanation_type=self.explanation_type
            )

        if self.verbose:
            print_unbuffered(f'{self.explanation_type}s were calculated in {self.__repr__()}')

        return out

    def __iter__(self) -> ExplainerArray:
        self.current_i = 0
        return self

    def __next__(self) -> Explainer:
        if self.sub_calculators is None:
            raise StopIteration
        if self.current_i < len(self.sub_calculators):
            out = self.sub_calculators[self.current_i]
            self.current_i += 1
            return out
        else:
            raise StopIteration

    def __str__(self) -> str:
        length = 0
        if self.sub_calculators is not None:
            length = len(self.sub_calculators)
        return f"ExplainerArray with {length} calculators with {self.explanation_type} explanation type"

    def __repr__(self) -> str:
        length = 0
        if self.sub_calculators is not None:
            length = len(self.sub_calculators)
        return f"<ExplainerArray with {length} calculators with {self.explanation_type} explanation type>"



import dalex as dx
import numpy as np
import pandas as pd
import warnings
import sys
import os
import multiprocessing

from typing import List, Optional, Literal

from edgaro.data.dataset import Dataset
from edgaro.model.model import Model
from edgaro.explain.explainer_result import Explanation, ModelProfileExplanation, Curve, ModelPartsExplanation
from edgaro.base.utils import print_unbuffered


def _predict_func(model, data):
    if isinstance(data, np.ndarray):
        data = pd.DataFrame(data)
    data.columns = model.get_test_dataset().data.columns
    return np.array(model.predict_proba(Dataset('', data, None)).target)


class Explainer:
    """
    The class defines Explainer for a Model object - it allows to calculate PDP [1]_ or ALE [2]_ curves.

    Parameters
    ----------
    model : Model
        A Model object to calculate explanations on.
    N : int, optional, default=None
        Number of observations that will be sampled from the test Dataset before the calculation of profiles
        (PDP/ALE curves). None means all data.
    explanation_type : {'PDP', 'ALE', 'VI'}, default='PDP'
        An explanation type to be calculated (`PDP` - Partial Dependence Profile, `ALE` - Accumulated Local Effects,
        `VI` - Variable Importance)
    verbose : bool, default=False
        Print messages during calculations.
    processes : int, default=1
        Number of processes for the calculation of explanations.
        If -1, it is replaced with the number of available CPU cores.
    random_state : int, optional, default=None
        Random state seed.
    B : int, optional, default=10
        Number of permutation rounds to perform on each variable - applicable only if explanation_type='VI'.

    Attributes
    ----------
    model : Model
        A Model object to calculate explanations on.
    name: str
        A name of the Explainer, by default it is a Model name.
    N : int, optional, default=None
        Number of observations that will be sampled from the test Dataset before the calculation of profiles
        (PDP/ALE curves). None means all data.
    explanation_type : {'PDP', 'ALE', 'VI'}
        An explanation type to be calculated.
    verbose : bool
        Print messages during calculations.
    explainer : dx.Explainer, optional
        An explainer object from `dalex` package.
    processes : int
        Number of processes for the calculation of explanations.
        If -1, it is replaced with the number of available CPU cores.
    random_state : int, optional
        Random state seed
    B : int, optional
        Number of permutation rounds to perform on each variable - applicable only if explanation_type='VI'.

    References
    ----------
    .. [1] https://ema.drwhy.ai/partialDependenceProfiles.html
    .. [2] https://ema.drwhy.ai/accumulatedLocalProfiles.html

    """

    def __init__(self, model: Model, N: Optional[int] = None, explanation_type: Literal['PDP', 'ALE', 'VI'] = 'PDP',
                 verbose: bool = False, processes: int = 1, random_state: Optional[int] = None,
                 B: Optional[int] = 10) -> None:
        self.model = model
        self.explainer = None
        self.name = model.name
        self.N = N
        self.explanation_type = explanation_type
        self.verbose = verbose
        self.processes = processes
        self.random_state = random_state
        self.B = B

        if self.processes == -1:
            self.processes = multiprocessing.cpu_count()

    def fit(self) -> None:
        """
        Fit the Explainer object and create an explainer attribute.
        """

        dataset = self.model.get_test_dataset()
        dataset = self.model.transform_target(dataset) # ??? VI residual

        if dataset is None:
            raise Exception('Error with dataset!')

        if dataset.target is None:
            raise Exception('Target data is not provided!')
        if dataset.data is None:
            raise Exception('Data in dataset is not provided!')

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            if not self.verbose:
                sys.stdout = open(os.devnull, 'w')

            self.explainer = dx.Explainer(self.model, dataset.data, dataset.target, label=dataset.name,
                                          verbose=self.verbose, predict_function=_predict_func)
            if not self.verbose:
                sys.stdout = sys.__stdout__

        if self.verbose:
            print_unbuffered(f'dalex explainer inside {self.__repr__()} was created with {dataset.name}')

    def transform(self, variables: Optional[List[str]] = None) -> Explanation:
        """
        Calculate the explanation.

        Parameters
        ----------
        variables : list[str], optional
            List of variables for which the explanation should be calculated.

        Returns
        -------
        Explanation
        """
        if self.explainer is None:
            raise Exception('Explainer was not fitted!')
        category_colnames_base = self.model.get_category_colnames()
        dict_output = {}

        if variables is None:
            variables = list(self.model.get_train_dataset().data.columns)

        if self.verbose:
            print_unbuffered(f'{self.explanation_type} is being calculated in {self.__repr__()} for '
                             f'{self.model.get_test_dataset().name}')

        if self.explanation_type == 'PDP' or self.explanation_type == 'ALE':
            explanation_type = 'partial' if self.explanation_type == 'PDP' else 'accumulated'

            self.__transform_curve_category(dict_output, variables, category_colnames_base, explanation_type)
            self.__transform_curve_other(dict_output, variables, category_colnames_base, explanation_type)

            if self.verbose:
                print_unbuffered(f'{self.explanation_type} was calculated calculated in {self.__repr__()} for '
                                 f'{self.model.get_test_dataset().name}')

            return ModelProfileExplanation(dict_output, self.name, self.model.get_category_colnames())
        elif self.explanation_type == 'VI':
            explanation_type = 'variable_importance'
            self.__transform_feature_importance(dict_output, variables, explanation_type)

            if self.verbose:
                print_unbuffered(f'{self.explanation_type} was calculated calculated in {self.__repr__()} for '
                                 f'{self.model.get_test_dataset().name}')

            return ModelPartsExplanation(dict_output, self.name)
        else:
            raise Exception('Wrong curve type!')

    def __transform_curve_category(self, dict_output, variables, category_colnames_base, explanation_type):
        category_colnames = list(set(variables).intersection(set(category_colnames_base)))
        if len(category_colnames) > 0:
            out_category = self.explainer.model_profile(verbose=False, variables=category_colnames,
                                                        variable_type='categorical',
                                                        N=self.N, type=explanation_type,
                                                        processes=self.processes, random_state=self.random_state)
            y = out_category.result['_yhat_']
            x = out_category.result['_x_']
            variable_names = out_category.result['_vname_'].unique()
            for i in range(len(variable_names)):
                dict_output[str(variable_names[i])] = Curve(x, y)

    def __transform_curve_other(self, dict_output, variables, category_colnames_base, explanation_type):
        other_colnames = list(set(variables).difference(set(category_colnames_base)))
        if len(other_colnames) > 0:
            out_others = self.explainer.model_profile(verbose=False, variables=other_colnames,
                                                      variable_type='numerical',
                                                      N=self.N, type=explanation_type,
                                                      processes=self.processes, random_state=self.random_state)
            variable_names = out_others.result['_vname_'].unique()
            y = out_others.result['_yhat_']
            x = out_others.result['_x_']
            length = y.size / len(variable_names)
            for i in range(len(variable_names)):
                lower = int(i * length)
                higher = int((i + 1) * length)
                dict_output[str(variable_names[i])] = Curve(x[lower:higher], y[lower:higher])

    def __transform_feature_importance(self, dict_output, variables, explanation_type):
        out = self.explainer.model_parts(verbose=False, variables=variables,
                                         N=self.N, B=self.B, type=explanation_type,
                                         processes=self.processes, random_state=self.random_state)
        out = out.result
        out = out[(out.variable != '_full_model_') & (out.variable != '_baseline_')]
        out = out[['variable', 'dropout_loss']]

        for index, row in out.iterrows():
            dict_output[row[0]] = row[1]

    def __str__(self) -> str:
        return f"Explainer for model {self.name} with {self.explanation_type}  explanation type"

    def __repr__(self) -> str:
        return f"<Explainer for model {self.name} with {self.explanation_type} explanation type>"



from __future__ import annotations

import openml
import os
import pandas as pd
import numpy as np

from typing import List, Union, Optional
from imblearn.datasets import fetch_datasets
from sklearn.model_selection import train_test_split

from edgaro.data.dataset import Dataset, DatasetFromOpenML
from edgaro.base.utils import print_unbuffered


class DatasetArray:
    """ Create a DatasetArray object

    This class creates a unified representation of an array of the Dataset objects, which can be further processed by
    other package classes.

    Parameters
    ----------
    datasets : list[Dataset, DatasetArray]
        The list of Dataset and DatasetArray objects.
    name : str, default='dataset_array'
        Name of the dataset array.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    name : str
        Name of the dataset array.
    datasets : list[Dataset, DatasetArray]
        The list of Dataset and DatasetArray objects.
    verbose : bool
        Print messages during calculations.
    """

    def __init__(self, datasets: List[Union[Dataset, DatasetArray]],
                 name: str = 'dataset_array', verbose: bool = False) -> None:
        keys = [df.name for df in datasets]
        if len(set(keys)) == len(keys):
            self.verbose = verbose
            self.datasets = datasets
            self.name = name

            if self.verbose:
                print_unbuffered(f'DatasetArray {self.__repr__()} created')
        else:
            raise Exception('Dataset names are not unique!')

    def __getitem__(self, key: Union[Union[str, int], List[Union[str, int]]]) -> Optional[Union[DatasetArray, Dataset]]:
        if isinstance(key, list):
            outs = [self.__getitem__(k) for k in key]
            outs = [o for o in outs if o is not None]
            out = DatasetArray(outs, self.name + "_subset")
            if len(out) == 0:
                return None
            else:
                return out
        elif isinstance(key, str):
            for df in self.datasets:
                if df.name == key:
                    return df
        elif isinstance(key, int):
            if key <= len(self.datasets):
                return self.datasets[key]
        return None

    def __len__(self) -> int:
        return len(self.datasets)

    def __eq__(self, other) -> bool:
        if not isinstance(other, DatasetArray):
            return False
        if not self.name == other.name:
            return False
        if not len(self) == len(other):
            return False
        for m in self.datasets:
            if m not in other.datasets:
                return False
        return True

    def __iter__(self) -> DatasetArray:
        self.current_i = 0
        return self

    def __next__(self) -> Union[Dataset, DatasetArray]:
        if self.current_i < len(self.datasets):
            out = self.datasets[self.current_i]
            self.current_i += 1
            return out
        else:
            raise StopIteration

    def train_test_split(self, test_size: float = 0.2, random_state: Optional[int] = None) -> None:
        """
        Split each of the Dataset objects into train and test datasets.

        Parameters
        ----------
        test_size : float, default=0.2
            The size of a train dataset.
        random_state : int, optional, default=None
            Random state seed.
        """
        for ds in self.datasets:
            ds.train_test_split(test_size, random_state)

        if self.verbose:
            print_unbuffered(f'DatasetArray {self.__repr__()} was train-test-split')

    @property
    def train(self) -> DatasetArray:
        """
        DatasetArray : the DatasetArray of train part of Dataset objects if the Dataset objects were train-test-split
        """
        out = [ds.train for ds in self.datasets]
        return DatasetArray(out, self.name + '_train')

    @property
    def test(self) -> DatasetArray:
        """
        DatasetArray : the DatasetArray of train part of Dataset objects if the Dataset objects were train-test-split
        """
        out = [ds.test for ds in self.datasets]
        return DatasetArray(out, self.name + '_test')

    def remove_nans(self, col_thresh: float = 0.9) -> None:
        """
        Remove rows with NaN values and columns containing almost only NaN values.

        Parameters
        ----------
        col_thresh : float, default=0.9
            The threshold of NaN values in columns above which a column should be dropped
        """
        for dataset in self.datasets:
            dataset.remove_nans(col_thresh=col_thresh)
        self.remove_empty_datasets()

    def remove_outliers(self, n_std: Union[float, int] = 3) -> None:
        """
        Remove outliers with NaN values and columns containing almost only NaN values.

        It is only applicable for continuous variables (that means not `category`, 'object' and 'int' type).

        Parameters
        ----------
        n_std : float, int, default=3
            Number of standard deviations.
            The observations that lies outside the range `column_mean +/- n_std*column_std` will be removed.
        """
        for dataset in self.datasets:
            dataset.remove_outliers(n_std=n_std)

    def remove_non_binary_target_datasets(self) -> None:
        """
        Remove Dataset objects which do not represent binary classification task.
        """
        for dataset in self.datasets:
            if isinstance(dataset, DatasetArray):
                dataset.remove_non_binary_target_datasets()

        # remove empty DatasetArrays
        self.datasets = [dataset for dataset in self.datasets if
                         not isinstance(dataset, DatasetArray) or len(dataset) > 0]

        # remove non-binary classification tasks
        self.datasets = [dataset
                         for dataset in self.datasets
                         if isinstance(dataset, DatasetArray) or (
                                 isinstance(dataset, Dataset) and dataset.check_binary_classification())
                         ]

        if self.verbose:
            print_unbuffered(f'Non binary datasets were removed from DatasetArray {self.__repr__()}')

    def remove_empty_datasets(self) -> None:
        """
        Remove empty Dataset objects.
        """
        for dataset in self.datasets:
            if isinstance(dataset, DatasetArray):
                dataset.remove_empty_datasets()

        # remove empty DatasetArrays
        self.datasets = [dataset for dataset in self.datasets if
                         not isinstance(dataset, DatasetArray) or len(dataset) > 0]

        # remove empty Datasets
        self.datasets = [dataset
                         for dataset in self.datasets
                         if isinstance(dataset, DatasetArray) or (
                                 isinstance(dataset, Dataset) and
                                 dataset.target is not None and len(dataset.target) != 0 and
                                 dataset.data is not None and len(dataset.data) != 0)
                         ]

        if self.verbose:
            print_unbuffered(f'Empty datasets were removed from DatasetArray {self.__repr__()}')

    def remove_categorical_and_ordinal_variables(self):
        """
        Remove categorical and ordinal variables.
        """
        for ds in self.datasets:
            ds.remove_categorical_and_ordinal_variables()

    def append(self, other: Union[Dataset, DatasetArray, List[Union[Dataset, DatasetArray]]]) -> None:
        """
        Append new object to an DatasetArray.

        Parameters
        ----------
        other : Dataset, DatasetArray, list[Dataset, DatasetArray]
            The object to be appended to the array.
        """
        if isinstance(other, list):
            self.datasets += other
        else:
            self.datasets.append(other)

    def head(self, n: int = 10):
        """
        Get first `n` rows of each of the Dataset objects.

        Parameters
        ----------
        n : int, default=10
            Number of rows.

        Returns
        -------
        DatasetArray
            A DatasetArray object with Dataset objects with `n` first rows.
        """
        return DatasetArray(
            datasets=[d.head(n) for d in self.datasets],
            name=self.name + '_head',
            verbose=self.verbose
        )

    def __str__(self) -> str:
        return ''.join([str(ds) + '\n' for ds in self.datasets])

    def __repr__(self) -> str:
        return f"<DatasetArray {self.name} with {len(self.datasets)} datasets>"


class DatasetArrayFromOpenMLSuite(DatasetArray):
    """
    Create a DatasetArray object from an `OpenML` suite.

    Before using this class, you should follow the procedure of configuring Authentication on the website
    `here <https://openml.github.io/openml-python/main/examples/20_basic/introduction_tutorial.html#sphx-glr-examples-20-basic-introduction-tutorial-py>`_.

    Otherwise, you should have your own API key for OpenML and pass it as a parameter.

    Parameters
    ----------
    suite_name : str, default = 'OpenML100'
        A task ID for a dataset in OpenML.
    apikey : str, optional, default=None
        An API key to OpenML (if you configured OpenML, you do not need to pass this parameter).
    name : str
        Name of the dataset array.
    verbose : bool, default=False
        Print messages during calculations.
    """

    def __init__(self, suite_name: str = 'OpenML100', apikey: Optional[str] = None,
                 name: str = 'dataset_array', verbose: bool = False) -> None:
        if openml.config.apikey == '':
            if apikey is None:
                raise Exception('API key is not available!')
            else:
                openml.config.apikey = apikey

        benchmark_suite = openml.study.get_suite(suite_name)

        if verbose:
            print_unbuffered(f'Benchmark suite data was downloaded for {suite_name}')

        dataset_array = []
        for i in benchmark_suite.data:
            try:
                ds = DatasetFromOpenML(task_id=i, apikey=apikey)
                dataset_array.append(ds)
            except openml.exceptions.OpenMLServerException:
                print(f'The dataset numer {i} was not downloaded due to the server exception!')

        self.__openml_name = benchmark_suite.name if 'name' in benchmark_suite.__dict__.keys() else ''
        self.__openml_description = benchmark_suite.description if 'description' in benchmark_suite.__dict__.keys() else ''

        super().__init__(datasets=dataset_array, name=name, verbose=verbose)

        if verbose:
            print_unbuffered(f'DatasetArray from OpenML benchmark suite {suite_name} was created')

    def openml_description(self) -> str:
        """
        Description of the suite from OpenML.

        Returns
        -------
        str
            The suite description.
        """
        return "Name: " + self.__openml_name + '\n' + 'Description: ' + '\n' + self.__openml_description


class DatasetArrayFromDirectory(DatasetArray):
    """ Create a DatasetArray object by loading `*.csv` and `*.npy` files.

    The `*.npy` files are the files, which contain numpy arrays or pickled objects. They are loaded using this
    `function <https://numpy.org/doc/stable/reference/generated/numpy.load.html>`_.

    The class assumes that the last column in each file is a target column.

    Parameters
    ----------
    path : str
        A path of a directory to load files from.
    name : str, default='dataset_array'
        Name of the dataset array.
    verbose : bool, default=False
        Print messages during calculations.
    """

    def __init__(self, path: str, name: str = 'dataset_array', verbose: bool = False) -> None:
        if not os.path.exists(path):
            raise Exception('The path does not exist!')
        if not os.path.isdir(path):
            raise Exception('The path argument is not a directory!')

        if verbose:
            print_unbuffered(f'The files from {path} are being loaded')

        dataset_array = []
        for filename in os.listdir(path):
            f = os.path.join(path, filename)
            if os.path.isfile(f):
                if f.lower().endswith('.csv') or f.lower().endswith('.npy'):
                    if f.lower().endswith('.csv'):
                        d = pd.read_csv(f)
                    elif f.lower().endswith('.npy'):
                        d = pd.DataFrame(np.load(f, allow_pickle=True))

                    y = d[d.columns[-1]]
                    X = d.drop(d.columns[-1], axis=1)
                    dataset_array.append(
                        Dataset(dataframe=X, target=y, name=os.path.splitext(filename)[0])
                    )

        if verbose:
            print_unbuffered(f'The files from {path} were loaded')

        super().__init__(datasets=dataset_array, name=name, verbose=verbose)


def load_benchmarking_set(apikey: Optional[str] = None, keep_categorical: bool = False, minimal_IR: float = 1.5,
                          minimal_n_rows: int = 1000, percent_categorical_to_remove: float = 0.75):
    """
    The function loads an example benchmarking set.

    Parameters
    ----------
    apikey : str, optional, default=None
        An API key to OpenML (if you configured OpenML, you do not need to pass this parameter).
        For details see `DatasetArrayFromOpenMLSuite` class documentation.
    keep_categorical : bool, default=False
        If True, the datasets will remain categorical variables.
    minimal_IR : float, default=1.5
        Minimal IR in the set.
    minimal_n_rows : int, default=1000
        Minimal number of rows in a Dataset.
    percent_categorical_to_remove : float, default=0.75
        Only applicable if keep_categorical=False; if categorical and nominal variables are above that number of
        all variables, the Dataset is removed.

    Returns
    -------
    DatasetArray
    """

    # Functions for changing some column types as `edgaro` assumes integers to be categorical

    def convert_to_float(ds_array, dataset_name, column_name):
        ds_array[dataset_name].data[column_name] = ds_array[dataset_name].data[column_name].astype('float64')

    def convert_columns_to_int(ds_array, dataset_name, cols):
        for col in cols:
            convert_to_float(ds_array, dataset_name, col)

    def convert_all_int_columns_to_float(ds_array, dataset_name):
        d = ds_array[dataset_name].data
        for col in list(d.select_dtypes(['uint8', 'int']).columns):
            convert_to_float(ds_array, dataset_name, col)

    exclude_global = ['nomao', 'Internet-Advertisements', 'isolet', 'sylva_agnostic', 'webpage', 'scene', 'protein_homo']

    # OpenML-CC18
    df_openml = DatasetArrayFromOpenMLSuite('OpenML-CC18', apikey=apikey)
    df_openml.remove_nans()
    df_openml.remove_non_binary_target_datasets()
    df_openml.remove_empty_datasets()
    stats = pd.DataFrame({'name': [], 'IR': [], 'nrow': [], 'ncol': []})

    for df in df_openml:
        tmp = pd.DataFrame({'name': [df.name], 'IR': [df.imbalance_ratio],
                            'nrow': [df.data.shape[0]], 'ncol': [df.data.shape[1]]})
        stats = pd.concat([stats, tmp])

    stats = stats[np.logical_and(stats.IR >= minimal_IR, stats.nrow >= minimal_n_rows)]
    stats = stats[np.logical_not(np.isin(np.array(stats.name), exclude_global))]
    df_openml = df_openml[list(stats.name)]

    if not keep_categorical:
        for i in range(len(df_openml)):
            df_openml[i].data = df_openml[i].data.convert_dtypes()

        n_cols = [ds.data.shape[1] for ds in df_openml]
        colnames = list(stats.name)
        df_openml.remove_categorical_and_ordinal_variables()
        n_cols_after = [ds.data.shape[1] for ds in df_openml]
        exclude = []

        for i in range(len(n_cols_after)):
            if n_cols_after[i] / n_cols[i] < 1 - percent_categorical_to_remove:
                exclude.append(colnames[i])

        for exc in exclude:
            colnames.remove(exc)

        df_openml = df_openml[colnames]
    else:
        convert_all_int_columns_to_float(df_openml, 'spambase')
        convert_all_int_columns_to_float(df_openml, 'qsar-biodeg')
        convert_all_int_columns_to_float(df_openml, 'credit-g')
        convert_all_int_columns_to_float(df_openml, 'adult')
        convert_all_int_columns_to_float(df_openml, 'kc1')
        convert_all_int_columns_to_float(df_openml, 'pc1')
        convert_all_int_columns_to_float(df_openml, 'pc3')
        convert_all_int_columns_to_float(df_openml, 'pc4')
        convert_all_int_columns_to_float(df_openml, 'bank-marketing')
        convert_columns_to_int(df_openml, 'churn', ['account_length', 'number_vmail_messages', 'total_day_calls',
                                                    'total_eve_calls', 'total_night_calls', 'total_intl_calls',
                                                    'number_customer_service_calls'])

    # OpenML-100
    df_openml2 = DatasetArrayFromOpenMLSuite('OpenML100', apikey=apikey)
    df_openml2.remove_nans()
    df_openml2.remove_non_binary_target_datasets()
    df_openml2.remove_empty_datasets()
    stats2 = pd.DataFrame({'name': [], 'IR': [], 'nrow': [], 'ncol': []})

    for df in df_openml2:
        tmp = pd.DataFrame({'name': [df.name], 'IR': [df.imbalance_ratio],
                            'nrow': [df.data.shape[0]], 'ncol': [df.data.shape[1]]})
        stats2 = pd.concat([stats2, tmp])

    stats2 = stats2[np.logical_and(stats2.IR >= minimal_IR, stats2.nrow >= minimal_n_rows)]
    stats2 = stats2[np.logical_not(stats2.name.isin(stats.name))]  # so as not to repeat datasets
    stats2 = stats2[np.logical_not(np.isin(np.array(stats2.name), exclude_global))]
    df_openml2 = df_openml2[list(stats2.name)]

    df_openml2['SpeedDating'].data = df_openml2['SpeedDating'].data[[col
                                                                     for col in df_openml2['SpeedDating'].data.columns
                                                                     if col == 'd_age' or not col.startswith('d_')]]

    if not keep_categorical:
        for i in range(len(df_openml2)):
            df_openml2[i].data = df_openml2[i].data.convert_dtypes()

        n_cols = [ds.data.shape[1] for ds in df_openml2]
        colnames = list(stats2.name)
        df_openml2.remove_categorical_and_ordinal_variables()
        n_cols_after = [ds.data.shape[1] for ds in df_openml2]
        exclude = []

        for i in range(len(n_cols_after)):
            if n_cols_after[i] / n_cols[i] < 1 - percent_categorical_to_remove:
                exclude.append(colnames[i])

        for exc in exclude:
            colnames.remove(exc)

        df_openml2 = df_openml2[colnames]
    else:
        convert_all_int_columns_to_float(df_openml2, 'steel-plates-fault')

    # imblearn
    datasets = fetch_datasets()
    stats3 = pd.DataFrame({'name': [], 'IR': [], 'nrow': [], 'ncol': []})
    df_tab = []

    for key in datasets.keys():
        if key != 'ozone_level':  # this dataset is already included in previous set
            ds = datasets[key]
            df = Dataset(dataframe=pd.DataFrame(ds.data), target=pd.Series(ds.target), name=ds.DESCR)
            df_tab.append(df)
            tmp = pd.DataFrame({'name': [df.name], 'IR': [df.imbalance_ratio],
                                'nrow': [df.data.shape[0]], 'ncol': [df.data.shape[1]]})
            stats3 = pd.concat([stats3, tmp])

    stats3 = stats3[np.logical_and(stats3.IR >= minimal_IR, stats3.nrow >= minimal_n_rows)]
    stats3 = stats3[np.logical_not(stats3.name.isin(stats.name))]  # so as not to repeat datasets
    stats3 = stats3[np.logical_not(stats3.name.isin(stats2.name))]  # so as not to repeat datasets
    exclude = ['optical_digits', 'satimage', 'pen_digits',
               'letter_img']  # excluding datasets made from images since they are not real tabular data.
    stats3 = stats3[np.logical_not(np.isin(np.array(stats3.name), exclude))]
    stats3 = stats3[np.logical_not(np.isin(np.array(stats3.name), exclude_global))]
    df_tab = [d for d in df_tab if d.name in list(stats3.name)]

    # `protein_homo` dataset is very big and after balancing, for example to `IR=1`, it would have gigantic sizes
    # therefore, only subset of this dataset will be included in research
    # for df in df_tab:
    #     if df.name == 'protein_homo':
    #         X, _, y, _ = train_test_split(df.data, df.target, train_size=30000, stratify=df.target, random_state=42)
    #         df.data = X
    #         df.target = y

    if not keep_categorical:
        for i in range(len(df_tab)):
            df_tab[i].data = df_tab[i].data.convert_dtypes()

        n_cols = [ds.data.shape[1] for ds in df_tab]
        colnames = list(stats3.name)

        for i in range(len(df_tab)):
            df_tab[i].remove_categorical_and_ordinal_variables()

        n_cols_after = [ds.data.shape[1] for ds in df_tab]
        exclude = []

        for i in range(len(n_cols_after)):
            if n_cols_after[i] / n_cols[i] < 1 - percent_categorical_to_remove:
                exclude.append(colnames[i])

        for exc in exclude:
            colnames.remove(exc)

        df_tab = [d for d in df_tab if d.name in colnames]

    out = DatasetArray(
        datasets=df_openml.datasets + df_openml2.datasets + df_tab,
        name='benchmarking_set',
        verbose=False
    )

    for d in out:
        d.data = d.data.astype(float)

    return out



from __future__ import annotations

import pandas as pd

from typing import Optional, List, Dict, Any, Union, Callable

from edgaro.base.base_transformer_array import BaseTransformerArray
from edgaro.model.model import Model
from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.base.utils import print_unbuffered


class ModelArray(BaseTransformerArray):
    """
    Create a class to train Models for each of the Dataset in DatasetArray.

    Parameters
    ----------
    base_model : Model
        The object defining the basic Model training procedure.
        The base_model object has to be clean - it cannot be fitted earlier.
    parameters : list[list, Dict[str, Any]]], optional
        The list of parameters for base_model. If the object is used for a DatasetArray object,
        the parameter list should be nested. For details, see Examples section.
    name : str
        A name of the ModelArray.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    name : str
        A name of the ModelArray.
    verbose : bool
        Print messages during calculations.

    Examples
    ----------
    Example 1

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.model.model import RandomForest
    >>> from edgaro.model.model_array import ModelArray
    >>> df = Dataset(name_1, df_1, target_1)
    >>> params = [{'n_estimators': 20}]
    >>> model = RandomForest()
    >>> array = ModelArray(model, parameters=params)
    >>> array.fit(df)
    >>> array.predict(df)

    Example 2

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.model.model import RandomForest
    >>> from edgaro.model.model_array import ModelArray
    >>> df = DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)])
    >>> params = [[{'n_estimators': 20}] for _ in range(len(df)) ]
    >>> model = RandomForest()
    >>> array = ModelArray(model, parameters=params)
    >>> array.fit(df)
    >>> array.predict(df)

    Example 3

    >>> from test.resources.objects import *
    >>> from edgaro.data.dataset import Dataset
    >>> from edgaro.data.dataset_array import DatasetArray
    >>> from edgaro.model.model import RandomForest
    >>> from edgaro.model.model_array import ModelArray
    >>> df = DatasetArray([
    ...         Dataset(name_2, df_1, target_1),
    ...         DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)])
    ... ])
    >>> params = [[{'n_estimators': 20}], [{'n_estimators': 10}, {'n_estimators': 30}]]
    >>> model = RandomForest()
    >>> array = ModelArray(model, parameters=params)
    >>> array.fit(df)
    >>> array.predict(df)

    """

    def __init__(self, base_model: Model, parameters: Optional[List[Union[Dict[str, Any], List]]] = None,
                 name: str = '', verbose: bool = False) -> None:
        super().__init__(base_transformer=base_model, parameters=parameters)
        self.name = name
        self.verbose = verbose

    def fit(self, dataset: Union[Dataset, DatasetArray]) -> None:
        """
        Fit the ModelArray.

        The fitting process includes encoding the categorical variables with OrdinalEncoder (from scikit-learn library)
        and target encoding (custom encoding, the minority class is encoded as 1, the majority class as 0).

        The method assumes that categorical variables, which has to be encoded,
        are one of the types: 'category', 'object'.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            The object to fit Model on.

        """
        if self.name == '':
            self.name = dataset.name

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} is being fitted with {dataset.name}')

        super().fit(dataset)
        self.__fix_classes(dataset)

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} was fitted with {dataset.name}')

    def predict(self, dataset: Union[Dataset, DatasetArray]) -> Union[Dataset, DatasetArray]:
        """
        Predict the class for a Dataset/DatasetArray object.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            A Dataset/DatasetArray object to make predictions on.

        Returns
        -------
        Dataset, DatasetArray

        """
        out = super().transform(dataset)

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} predicted on {dataset.name}')

        return out

    @staticmethod
    def __set_to_probs(inp: Union[List, Model, ModelArray]):
        if isinstance(inp, list):
            for i in inp:
                ModelArray.__set_to_probs(i)
        else:
            inp.set_transform_to_probabilities()

    @staticmethod
    def __set_to_class(inp: Union[List, Model, ModelArray]):
        if isinstance(inp, list):
            for i in inp:
                ModelArray.__set_to_class(i)
        else:
            inp.set_transform_to_classes()

    def predict_proba(self, dataset: Union[Dataset, DatasetArray]) -> Union[Dataset, DatasetArray]:
        """
        Predict the probability of class `1` for a Dataset/DatasetArray object.

        Parameters
        ----------
        dataset : Dataset, DatasetArray
            A Dataset/DatasetArray object to make predictions on.

        Returns
        -------
        Dataset, DatasetArray

        """
        ModelArray.__set_to_probs(self.get_models())
        out = super().transform(dataset)
        ModelArray.__set_to_class(self.get_models())

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} predicted probabilities on {dataset.name}')

        return out

    def get_models(self) -> List[Union[Model, ModelArray, List[Union[Model, ModelArray]]]]:
        """
        All the Model/ModelArray objects used by this object.

        Returns
        -------
        list[Model, ModelArray, list]
        """
        return self.transformers

    def set_transform_to_probabilities(self) -> None:
        """
        Make `transform` function return probabilities.
        """
        ModelArray.__set_to_probs(self.get_models())

    def set_transform_to_classes(self) -> None:
        """
        Make `transform` function return classes..
        """
        ModelArray.__set_to_class(self.get_models())

    def __fix_classes(self, dataset: Union[Dataset, DatasetArray]) -> None:
        for i in range(len(self.transformers)):
            self.transformers[i] = self.__base_transformer_array_to_model_array(
                self.transformers[i],
                dataset[i].name if isinstance(dataset, DatasetArray) else dataset.name
            )
            if isinstance(self.transformers[i], BaseTransformerArray):
                self.transformers[i].__fix_classes(dataset if isinstance(dataset, Dataset) else dataset[i])

    def __base_transformer_array_to_model_array(self, base: Union[Model, ModelArray, BaseTransformerArray],
                                                name: str) -> Union[Model, ModelArray]:
        if isinstance(base, Model):
            return base
        elif not isinstance(base, ModelArray) and isinstance(base, BaseTransformerArray):
            out = ModelArray(base_model=self.base_transformer)
            out.__class__ = self.__class__
            for key, val in base.__dict__.items():
                out.__dict__[key] = val
            out.name = name

            return out
        else:
            return base

    def evaluate(self, metrics_output_class: Optional[List[Callable[[pd.Series, pd.Series], float]]] = None,
                 metrics_output_probabilities: Optional[List[Callable[[pd.Series, pd.Series], float]]] = None,
                 ds: Optional[Union[Dataset, DatasetArray]] = None) -> pd.DataFrame:
        """
        Evaluate model.

        Parameters
        ----------
        metrics_output_class : list[Callable[[pd.Series, pd.Series], float]], optional, default=None
            List of functions to calculate metrics on predicted classes. If None is passed, accuracy, balanced accuracy,
            precision, recall, specificity, f1, f1_weighted, geometric mean score are used.
        metrics_output_probabilities : list[Callable[[pd.Series, pd.Series], float]], optional, default=None
            List of functions to calculate metrics on predicted probabilities. If None is passed, ROC AUC is used.
        ds : Dataset, DatasetArray, optional, default=None
            A Dataset/DatasetArray object to calculate metric on. If None is passed,
            test Dataset/DatasetArray from fitting is used.

        Returns
        -------
        pd.DataFrame
        """

        def _evaluate(mod, ds_, out_in):
            m = mod
            data = ds_ if isinstance(ds_, Dataset) else None
            eval_model = m.evaluate(metrics_output_class=metrics_output_class,
                                    metrics_output_probabilities=metrics_output_probabilities, ds=data)
            eval_model['model'] = m.name
            return pd.concat([out_in, eval_model])

        def _eval_all(mod, ds_, out_in):
            out_out = None
            if isinstance(mod, Model):
                out_out = _evaluate(mod, ds_, out_in)
            elif isinstance(mod, ModelArray):
                tmp = []
                for j in range(len(mod.get_models())):
                    m = mod.get_models()[j]
                    if isinstance(ds_, DatasetArray):
                        tmp.append(_eval_all(m, ds_[j], out_out))
                    else:
                        tmp.append(_eval_all(m, ds_, out_in))
                tmp.append(out_out)
                out_out = pd.concat(tmp)
            elif isinstance(mod, list):
                tmp = []
                for j in range(len(mod)):
                    if isinstance(ds_, DatasetArray):
                        tmp.append(_eval_all(mod[j], ds_[j], out_out))
                    else:
                        tmp.append(_eval_all(mod[j], ds_, out_out))
                tmp.append(out_out)
                out_out = pd.concat(tmp)
            return out_out

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} is being evaluated')

        out = pd.DataFrame({'model': [], 'metric': [], 'value': []})
        out = _eval_all(self.get_models(), ds, out)
        out = out[['model', 'metric', 'value']]
        out = out.reset_index(drop=True)

        if self.verbose:
            print_unbuffered(f'ModelArray {self.__repr__()} was evaluated')

        return out

    @property
    def transformers(self) -> List[Union[Model, ModelArray, List]]:
        """
        All the Model objects used by this object.

        Returns
        -------
        list[Model, ModelArray, list]
        """
        return super().transformers

    @transformers.setter
    def transformers(self, val: List[Union[Model, ModelArray, List]]) -> None:
        BaseTransformerArray.transformers.fset(self, val)

    @property
    def base_transformer(self) -> Model:
        """
        Base transformers for creation of this object.

        Returns
        -------
        Model
        """
        out = super().base_transformer
        if isinstance(out, Model):
            return out
        else:
            raise Exception('Wrong base_transformer attribute')

    @base_transformer.setter
    def base_transformer(self, val: Model) -> None:
        BaseTransformerArray.base_transformer.fset(self, val)

    def __str__(self) -> str:
        return f"ModelArray {self.name} with {len(self.get_models())} models"

    def __repr__(self) -> str:
        return f"<ModelArray {self.name} with {len(self.get_models())} models>"



import pandas as pd
import pytest

from copy import deepcopy

from edgaro.balancing.transformer import SMOTE
from edgaro.base.base_transformer import BaseTransformer
from edgaro.data.dataset import Dataset
from edgaro.data.dataset_array import DatasetArray
from edgaro.balancing.transformer_array import TransformerArray
from edgaro.balancing.nested_transformer import BasicAutomaticTransformer, ExtensionAutomaticTransformer, \
    AutomaticTransformer, NestedAutomaticTransformer

from .resources.objects import *


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_automatic_transformer_array(ds, n):
    try:
        array = BasicAutomaticTransformer(n_per_method=n)
        array.fit(ds)
        array.transform(ds)
        str(array)
        repr(array)
        str(super(NestedAutomaticTransformer, array))
        repr(super(NestedAutomaticTransformer, array))
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
@pytest.mark.parametrize('suffix', [
    '_suffix_1',
    '_example_sufix'
])
def test_automatic_transformer_array_suffix(ds, n, suffix):
    array = BasicAutomaticTransformer(n_per_method=n, result_array_sufix=suffix)

    assert len(array.get_dataset_suffixes()) == 0

    array.fit(ds)
    out = array.transform(ds)
    assert out.name == ds.name + suffix

    assert out[0].name.endswith('UNDERSAMPLING__Random')
    assert out[1].name.endswith('OVERSAMPLING__Random')
    assert out[2].name.endswith('OVERSAMPLING__SMOTE')

    try:
        out = array.get_dataset_suffixes()
    except (Exception,):
        assert False

    assert len(out) > 0


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_automatic_transformer_array_shape(ds, n):
    array = BasicAutomaticTransformer(n_per_method=n)
    array.fit(ds)
    out = array.transform(ds)

    assert len(out) == 3
    for d in out:
        assert len(d) == n


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
@pytest.mark.parametrize('min_samples_to_modify, n, expected_n', [
    (1, 5, 5),
    (2, 5, 5),
    (3, 5, 4),
    (4, 5, 3),
    (6, 5, 3),
    (7, 5, 2),
    (10, 5, 2),
    (20, 5, 1)
])
def test_automatic_transformer_array_shape_with_min_samples_to_modify(ds, min_samples_to_modify, n, expected_n):
    array = BasicAutomaticTransformer(n_per_method=n, min_samples_to_modify=min_samples_to_modify)
    array.fit(ds)
    out = array.transform(ds)

    assert len(out) == 3
    for d in out:
        assert len(d) == expected_n


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_automatic_transformer_array_was_fitted(ds, n):
    array = BasicAutomaticTransformer(n_per_method=n)
    assert not array.was_fitted
    array.fit(ds)
    assert array.was_fitted


@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
            Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))])])
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_array_of_automatic_transformer_array(ds, n):
    try:
        transformer = BasicAutomaticTransformer(n_per_method=n)
        array = TransformerArray(transformer)
        array.fit(ds)
        array.transform(ds)
        str(array)
        repr(array)
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))])
])
@pytest.mark.parametrize('n', [
    2, 3
])
@pytest.mark.parametrize('suffix', [
    '_suffix_1',
    '_example_sufix'
])
def test_array_of_automatic_transformer_array_suffix(ds, n, suffix):
    transformer = BasicAutomaticTransformer(n_per_method=n)
    array = TransformerArray(transformer, result_array_sufix=suffix, allow_dataset_array_sufix_change=False)
    array.fit(ds)
    out = array.transform(ds)

    assert out.name == ds.name + suffix

    for d in out:
        assert d[0].name.endswith('UNDERSAMPLING__Random')
        assert d[1].name.endswith('OVERSAMPLING__Random')
        assert d[2].name.endswith('OVERSAMPLING__SMOTE')


@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))])
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_array_of_automatic_transformer_array_was_fitted(ds, n):
    transformer = BasicAutomaticTransformer(n_per_method=n)
    array = TransformerArray(transformer)
    assert not array.was_fitted
    array.fit(ds)
    assert array.was_fitted


@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))]),
    DatasetArray([
        Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))])
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_array_of_automatic_transformer_array_shape(ds, n):
    transformer = BasicAutomaticTransformer(n_per_method=n)
    array = TransformerArray(transformer)
    array.fit(ds)
    out = array.transform(ds)

    assert len(out) == len(ds)
    for trans_for_ds in out:
        assert len(trans_for_ds) == 3
        for trans_for_ir in trans_for_ds:
            assert len(trans_for_ir) == n


@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1),
    Dataset(name_2, df_2, target_2)
])
def test_transform_empty(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    assert array.transform(ds) == DatasetArray([])


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
def test_transform_keep_original(ds):
    array = BasicAutomaticTransformer(n_per_method=2, keep_original_dataset=True)
    array.fit(ds)
    transform_true = array.transform(ds)

    array.keep_original_dataset = False
    transform_false = array.transform(ds)

    assert len(transform_true) - 1 == len(transform_false)


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
def test_output_verbose(ds, capsys):
    array = BasicAutomaticTransformer(n_per_method=2, verbose=True)
    array.fit(ds)
    array.transform(ds)

    captured = capsys.readouterr()
    assert f'AutomaticTransformerArray {array.__repr__()} was fitted with {ds.name}' in captured.out
    assert f'AutomaticTransformerArray {array.__repr__()} transformed {ds.name}' in captured.out


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
def test_get_set_transformers(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    assert array.transformers == []
    assert len(array) == 0
    array.fit(ds)
    array.transform(ds)
    assert len(array.transformers) > 0
    assert len(array) > 0

    with pytest.raises(Exception):
        array.transformers = deepcopy(array.transformers)

    array1 = BasicAutomaticTransformer(n_per_method=2)
    try:
        array1.transformers = deepcopy(array.transformers)
    except (Exception,):
        assert False

    try:
        array1.transform(ds)
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1)
])
def test_base_transformer_get_set(ds):
    array = BasicAutomaticTransformer(n_per_method=2)

    assert isinstance(array.base_transformer, list)

    try:
        ar = deepcopy(array)
        ar.base_transformer = TransformerArray(SMOTE())
        ar.fit(ds)
    except (Exception,):
        assert False

    with pytest.raises(Exception):
        ar = deepcopy(array)
        tmp = TransformerArray(SMOTE())
        tmp.__class__ = BaseTransformer
        ar.base_transformer = tmp
        x = ar.base_transformer


@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1)
])
def test_base_transformer_get_set_none(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    ar = deepcopy(array)
    ar.base_transformer = None
    assert ar.base_transformer == []


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_2 for _ in range(5)]), pd.concat([target_2 for _ in range(5)]))
])
def test_base_transformer_get_set_fail_fitted(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    ar = deepcopy(array)
    ar.fit(ds)
    with pytest.raises(Exception):
        ar.base_transformer = None


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)]))
])
def test_get_item(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    assert array[0] is None
    array.fit(ds)

    try:
        x = array[0]
        x = array[1]
        x = array[[1, 2]]
    except (Exception,):
        assert False

    assert array[123444] is None
    assert array[1.23] is None


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_2 for _ in range(5)]), pd.concat([target_2 for _ in range(5)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_extension_automatic_transformer_array(ds, n):
    try:
        array = ExtensionAutomaticTransformer(n_per_method=n)
        str(array)
        repr(array)
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_2 for _ in range(5)]), pd.concat([target_2 for _ in range(5)]))
])
@pytest.mark.parametrize('n', [
    2, 3
])
def test_extension_automatic_transformer_array(ds, n):
    try:
        array = AutomaticTransformer(n_per_method=n)
        str(array)
        repr(array)
    except (Exception,):
        assert False


@pytest.mark.parametrize('ds', [
    Dataset(name_1, pd.concat([df_2 for _ in range(5)]), pd.concat([target_2 for _ in range(5)]))
])
def test_get_set_params(ds):
    array = BasicAutomaticTransformer(n_per_method=2)
    array.fit(ds)

    try:
        param = array.get_params()
    except (Exception,):
        assert False

    assert isinstance(param, list)

    try:
        array.set_params(IR=[[1.5, 1] for _ in range(3)])
    except (Exception,):
        assert False




from __future__ import annotations

import openml
import pandas as pd
import numpy as np

from copy import deepcopy
from typing import Optional, Union
from pandas_profiling import ProfileReport
from sklearn.model_selection import train_test_split
from imblearn.datasets import fetch_datasets

from edgaro.base.utils import print_unbuffered


class Dataset:
    """ Create a Dataset object

    This class creates a unified representation of a dataset, which can be further processed by other package classes.

    Only one of `dataframe` and `target` parameters is required.

    Parameters
    ----------
    name : str
        Name of the dataset.
    dataframe : pd.DataFrame, optional
        The variables (predictors) in a dataset.
    target : pd.Series, optional
        The target in a dataset.
    verbose : bool, default=False
        Print messages during calculations.

    Attributes
    ----------
    name : str
        Name of the dataset.
    verbose : bool
        Print messages during calculations.
    majority_class_label : str, optional, default=None
        The label of the majority class. It is recommended to set this attribute; otherwise, it will be guessed while
        encoding in the `model` module. The guess may be wrong only if the dataset is balanced - consequently,
        experiment results may be wrong.
    """

    def __init__(self, name: str, dataframe: Optional[pd.DataFrame], target: Optional[pd.Series],
                 verbose: bool = False) -> None:
        if dataframe is not None and target is not None:
            if dataframe.shape[0] != target.shape[0]:
                raise Exception('Dataframe and target have different number of rows!')

        self.name = name
        self.verbose = verbose
        self.__data = dataframe
        self.__target = target

        self.__train_dataset: Optional[Dataset] = None
        self.__test_dataset: Optional[Dataset] = None

        self.majority_class_label = None

        if self.verbose:
            print_unbuffered(f'Dataset {self.__repr__()} created')

    @property
    def data(self) -> Optional[pd.DataFrame]:
        """
        pd.DataFrame, optional : The variables (predictors) in a dataset.
        """
        if self.__data is not None:
            return self.__data
        else:
            if self.__train_dataset is not None and self.__test_dataset is not None and \
                    self.__train_dataset.data is not None and self.__test_dataset.data is not None:
                return pd.concat([self.__train_dataset.data, self.__test_dataset.data])
            else:
                return None

    @data.setter
    def data(self, val: Optional[pd.DataFrame]) -> None:
        if self.__train_dataset is None and self.__test_dataset is None:
            self.__data = val
        else:
            raise Exception('Data cannot be set since the dataset was train-test-split!')

    @property
    def target(self) -> Optional[pd.Series]:
        """
        pd.Series, optional : The target in a dataset.
        """
        if self.__target is not None:
            return self.__target
        else:
            if self.__train_dataset is not None and self.__test_dataset is not None and \
                    self.__train_dataset.target is not None and self.__test_dataset.target is not None:
                return pd.concat([self.__train_dataset.target, self.__test_dataset.target])
            else:
                return None

    @target.setter
    def target(self, val: Optional[pd.Series]) -> None:
        if self.__train_dataset is None and self.__test_dataset is None:
            self.__target = val
        else:
            raise Exception('Target cannot be set since the dataset was train-test-split!')

    @property
    def train(self) -> Dataset:
        """
        Dataset : the train dataset if the Dataset object was train-test-split
        """
        if self.__train_dataset is not None:
            return self.__train_dataset
        else:
            raise Exception('The dataset as not train-test-split!')

    @property
    def test(self) -> Dataset:
        """
        Dataset : the test dataset if the Dataset object was train-test-split
        """
        if self.__test_dataset is not None:
            return self.__test_dataset
        else:
            raise Exception('The dataset as not train-test-split!')

    @property
    def was_split(self) -> bool:
        """
        bool : the information whether the Dataset object was train-test-split
        """
        return self.__train_dataset is not None and self.__test_dataset is not None

    def train_test_split(self, test_size: float = 0.2, random_state: Optional[int] = None) -> None:
        """
        Split the object into train and test datasets.

        Parameters
        ----------
        test_size : float, default=0.2
            The size of a train dataset.
        random_state : int, optional, default=None
            Random state seed.
        """
        if self.was_split:
            raise Exception('The dataset has already been train-test-split!')
        X_train, X_test, y_train, y_test = train_test_split(deepcopy(self.__data), deepcopy(self.__target),
                                                            test_size=test_size,
                                                            random_state=random_state, stratify=self.__target)
        self.__train_dataset = Dataset(self.name + '_train', X_train, y_train)
        self.__test_dataset = Dataset(self.name + '_test', X_test, y_test)

        self.__data = None
        self.__target = None

        if self.verbose:
            print_unbuffered(f'Dataset {self.__repr__()} was train-test-split')

    def custom_train_test_split(self, train: Dataset, test: Dataset) -> None:
        """
        Set custom train and set dataset.

        Parameters
        ----------
        train : Dataset
            The train dataset.
        test : Dataset
            The test dataset.
        """
        self.__train_dataset = train
        self.__test_dataset = test

        self.__data = None
        self.__target = None

    def check_binary_classification(self) -> bool:
        """
        The information whether the Dataset object contains binary classification data

        Returns:
            bool
        """
        if self.target is not None:
            unique = np.unique(self.target)
            if len(unique) > 2:
                return False
            return True
        else:
            return False

    def generate_report(self, output_path: Optional[str] = None, show_jupyter: bool = False,
                        minimal: bool = False) -> None:
        """
        Generate a report using `pandas_profiling` tool.

        Parameters
        ----------
        output_path : str, optional, default=None
            The path to save the generated report.
        show_jupyter : bool, default=False
            If set to `True`, the report will be displayed as a Jupyter Notebook IFrame.
        minimal : bool, default=False
            Turn off the most expensive computations.
        """
        if self.data is None and self.target is None:
            raise Exception('Both data and target are None!')

        if self.target is None:
            data = self.data
        elif self.data is None:
            data = pd.DataFrame(self.target)
        else:
            data = self.data.assign(target=self.target)

        profile = ProfileReport(data, title='Pandas Profiling Report for ' + self.name, minimal=minimal,
                                progress_bar=False)

        if output_path is not None:
            profile.to_file(output_file=output_path)
        if show_jupyter:
            profile.to_notebook_iframe()

    @property
    def imbalance_ratio(self) -> float:
        """
        float : Imbalance Ratio of the Dataset; is the ratio of the majority class to the minority class
        """
        if self.target is None:
            return float(0)
        names, counts = np.unique(self.target, return_counts=True)
        if len(names) == 1:
            return float(0)
        elif len(names) > 2:
            raise Exception('Target has too many classes for binary classification!')
        else:
            return float(max(counts) / min(counts))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Dataset):
            return False
        if self.name != other.name:
            return False
        if self.data is not None and not self.data.equals(other.data):
            return False
        if self.target is not None and not self.target.equals(other.target):
            return False
        return True

    def remove_nans(self, col_thresh: float = 0.9) -> None:
        """
        Remove rows with NaN values and columns containing almost only NaN values.

        Parameters
        ----------
        col_thresh : float, default=0.9
            The threshold of NaN values in columns above which a column should be dropped
        """
        if self.data is not None:
            nans = self.data.isnull().sum() / self.data.shape[0]
            nans = list(nans[nans > col_thresh].index)
            self.data.drop(nans, axis=1, inplace=True)

            nans = self.data.isna().any(axis=1)
            nans = list(nans[nans].index)
            self.data.drop(nans, axis=0, inplace=True)
            if self.target is not None:
                self.target.drop(nans, axis=0, inplace=True)

            nans = self.target.isna()
            nans = list(nans[nans].index)
            self.data.drop(nans, axis=0, inplace=True)
            if self.target is not None:
                self.target.drop(nans, axis=0, inplace=True)

    def remove_outliers(self, n_std: Union[float, int] = 3) -> None:
        """
        Remove outliers with NaN values and columns containing almost only NaN values.

        It is only applicable for continuous variables (that means not `category`, 'object' and 'int' type).

        Parameters
        ----------
        n_std : float, int, default=3
            Number of standard deviations.
            The observations that lies outside the range `column_mean +/- n_std*column_std` will be removed.
        """
        if self.data is not None and self.target is not None:
            categorical_columns = list(self.data.select_dtypes(include=['category', 'object', 'int']))
            numerical_columns = list(set(self.data.columns).difference(categorical_columns))

            for col in numerical_columns:
                mean = self.data[col].mean()
                std = self.data[col].std()
                index = np.logical_and(self.data[col] <= mean + (n_std * std), (self.data[col] >= mean - (n_std * std)))
                self.data = self.data[index]
                self.target = self.target[index]

    def remove_categorical_and_ordinal_variables(self):
        """
        Remove categorical and ordinal variables.
        """
        self.data = self.data.select_dtypes(exclude=["category", "uint8", "int64"])

    def head(self, n: int = 10):
        """
        Get first `n` rows of the Dataset.

        Parameters
        ----------
        n : int, default=10
            Number of rows.

        Returns
        -------
        Dataset
            A Dataset object with `n` first rows.
        """
        new_data = self.data.head(n)
        new_target = self.target.head(n)
        return Dataset(
            name=self.name + '_head',
            dataframe=new_data,
            target=new_target,
            verbose=self.verbose
        )

    def __str__(self) -> str:
        out = f"Name: {self.name}"
        if self.data is not None:
            out += f"Dataset: \n{self.data.head()}"
        if self.target is not None:
            out += f"Target: \n{self.target.head()}"
        if self.check_binary_classification():
            out += f"Imbalance ratio: \n{self.imbalance_ratio}"
        return out

    def __repr__(self) -> str:
        return f"<Dataset {self.name}>"


class DatasetFromCSV(Dataset):
    """ Create a Dataset object from a `*.csv` file

    If`target` parameter is `None`, the target is the last column in file.

    Parameters
    ----------
    name : str, default='dataset'
        Name of the dataset.
    path : str
        The path to `*.csv` fiole.
    target : str, optional, default=None
        The name of a column in the Dataset that is a target.
    verbose : bool, default=False
        Print messages during calculations.
    *args : tuple, optional
        Additional arguments to pd.read_csv function.
    **kwargs : dict, optional
        Additional arguments to pd.read_csv function.
    """

    def __init__(self, path: str, target: Optional[str] = None, name: str = 'dataset',
                 verbose: bool = False, *args, **kwargs) -> None:
        X = pd.read_csv(path, *args, **kwargs)

        if verbose:
            print_unbuffered(f'Data from {path} file is loaded')

        if target is None:
            y = X.iloc[:, -1]
            target = X.columns[-1]
        else:
            y = X[target]
        y = pd.Series(y, name='target')
        X = X.drop([target], axis=1)
        super().__init__(name=name, dataframe=X, target=y, verbose=verbose)


class DatasetFromOpenML(Dataset):
    """  Create a Dataset object from an `OpenML` dataset.

    Before using this class, you should follow the procedure of configuring Authentication on the website
    `here <https://openml.github.io/openml-python/main/examples/20_basic/introduction_tutorial.html#sphx-glr-examples-20-basic-introduction-tutorial-py>`_.

    Otherwise, you should have your own API key for OpenML and pass it as a parameter.

    Parameters
    ----------
    task_id : int
        A task ID for a dataset in OpenML.
    apikey : str, optional, default=None
        An API key to OpenML (if you configured OpenML, you do not need to pass this parameter).
    verbose : bool, default=False
        Print messages during calculations.
    """

    def __init__(self, task_id: int, apikey: Optional[str] = None, verbose: bool = False) -> None:
        if openml.config.apikey == '':
            if apikey is None:
                raise Exception('API key is not available!')
            else:
                openml.config.apikey = apikey

        data = openml.datasets.get_dataset(task_id)
        self.__openml_name = data.name if 'name' in data.__dict__.keys() else ''
        self.__openml_description = data.description if 'description' in data.__dict__.keys() else ''

        X, y, categorical_indicator, attribute_names = data.get_data(
            dataset_format='dataframe', target=data.default_target_attribute
        )

        if verbose:
            print_unbuffered(f'Dataset from OpenML with id {str(task_id)} was downloaded')

        X = pd.DataFrame(X, columns=attribute_names)
        y = pd.Series(y, name='target')

        for i in range(len(X.columns)):
            if categorical_indicator[i]:
                col = X.columns[i]
                col_type = X[col].dtype
                X[col] = np.array([col]).astype('category') if col_type not in ['category', 'object', 'int'] else X[col]

        super().__init__(name=data.name, dataframe=X, target=y, verbose=verbose)

    def openml_description(self) -> Optional[str]:
        """
        Description of the Dataset from OpenML.

        Returns
        -------
        str
            The Dataset description.
        """
        return "Name: " + self.name + '\n' + 'Description: ' + '\n' + self.__openml_description


def load_mammography():
    """
    The function loads an example dataset 'mammography'.

    Returns
    -------
    Dataset
    """

    mammography = fetch_datasets(filter_data=('mammography',))['mammography']
    X = mammography['data']
    y = mammography['target']
    name = mammography['DESCR']

    return Dataset(
        dataframe=pd.DataFrame(X),
        target=pd.Series(y),
        name=name
    )





import pytest

from imblearn.under_sampling import RandomUnderSampler
from sklearn.metrics import accuracy_score
from copy import deepcopy

from edgaro.data.dataset import DatasetFromOpenML, Dataset
from edgaro.balancing.transformer import TransformerFromIMBLEARN, RandomUnderSampler as RUS
from edgaro.balancing.transformer_array import TransformerArray
from edgaro.explain.explainer_result import ModelProfileExplanation
from edgaro.explain.explainer_result_array import ModelProfileExplanationArray
from edgaro.model.model import RandomForest
from edgaro.model.model_array import ModelArray
from edgaro.explain.explainer import Explainer
from edgaro.explain.explainer_array import ExplainerArray
from edgaro.data.dataset_array import DatasetArray

from .resources.objects import *


@pytest.fixture(
    scope='module',
    params=[
        Dataset(name_1, pd.concat([df_1 for _ in range(30)]), pd.concat([target_1 for _ in range(30)]))
    ]
)
def model(request):
    df = request.param
    df.remove_nans()
    transformator = TransformerFromIMBLEARN(RandomUnderSampler(sampling_strategy=1, random_state=42))
    transformator.fit(df)
    df_new = transformator.transform(df)

    rf = RandomForest(test_size=0.3, max_depth=1, n_estimators=1, random_state=42)
    rf.fit(df_new)
    return rf


def test_explainer(model, capsys):
    rf = deepcopy(model)

    try:
        pdp = Explainer(rf, N=10, verbose=True)
        pdp.fit()
        str(pdp)
        repr(pdp)
        pdp.transform()
    except (Exception,):
        assert False

    captured = capsys.readouterr()
    assert f'dalex explainer inside {pdp.__repr__()} was created with {name_1}' in captured.out
    assert f'{pdp.explanation_type} is being calculated in {pdp.__repr__()} for ' \
           f'{pdp.model.get_test_dataset().name}' in captured.out
    assert f'{pdp.explanation_type} was calculated calculated in {pdp.__repr__()} for ' \
           f'{pdp.model.get_test_dataset().name}' in captured.out


def test_transform_no_explainer(model, capsys):
    rf = deepcopy(model)
    pdp = Explainer(rf, N=10)

    with pytest.raises(Exception, ):
        pdp.transform()


def test_transform_ale(model, capsys):
    rf = deepcopy(model)
    pdp = Explainer(rf, N=10, explanation_type='ALE')
    pdp.fit()

    try:
        pdp.transform()
    except (Exception,):
        assert False


def test_transform_wrong_type(model, capsys):
    rf = deepcopy(model)
    pdp = Explainer(rf, N=10, explanation_type='XXX')

    with pytest.raises(Exception):
        pdp.transform()


@pytest.fixture(
    scope='module',
    params=[
        DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
            Dataset(name_2, pd.concat([df_1 for _ in range(20)]), pd.concat([target_1 for _ in range(20)])),
        ])
    ]
)
def model_array(request):
    df = request.param
    df.remove_nans()
    transformator = TransformerArray(
        TransformerFromIMBLEARN(RandomUnderSampler(sampling_strategy=1, random_state=42)))
    transformator.fit(df)
    df_new = transformator.transform(df)

    rf = ModelArray(RandomForest(test_size=0.3, max_depth=1, n_estimators=1, random_state=42))
    rf.fit(df_new)
    return rf


def test_explainer_array(model_array, capsys):
    rf = deepcopy(model_array)

    try:
        pdp = ExplainerArray(rf, N=10, verbose=True)
        pdp.fit()
        str(pdp)
        repr(pdp)
        pdp.transform()
    except (Exception,):
        assert False

    captured = capsys.readouterr()
    assert f'dalex explainers inside {pdp.__repr__()} were created' in captured.out
    assert f'{pdp.explanation_type}s are being calculated in {pdp.__repr__()}' in captured.out
    assert f'{pdp.explanation_type}s were calculated in {pdp.__repr__()}' in captured.out


def test_explainer_array_iterate(model_array):
    rf = deepcopy(model_array)
    pdp = ExplainerArray(rf, N=10)
    pdp.fit()
    try:
        for p in pdp:
            pass
    except (Exception,):
        assert False


def test_explainer_array_transform_not_fitted(model_array):
    rf = deepcopy(model_array)
    pdp = ExplainerArray(rf, N=10)

    with pytest.raises(Exception):
        pdp.transform()


@pytest.fixture(
    scope='module',
    params=[
        (Dataset(name_1, pd.concat([df_1 for _ in range(30)]), pd.concat([target_1 for _ in range(30)])), 'PDP'),
        (Dataset(name_1, pd.concat([df_1 for _ in range(30)]), pd.concat([target_1 for _ in range(30)])), 'ALE')
    ]
)
def pdp_transformed(request):
    df = request.param[0]
    df.remove_nans()
    transformator = TransformerFromIMBLEARN(RandomUnderSampler(sampling_strategy=1, random_state=42))
    transformator.fit(df)
    df_new = transformator.transform(df)

    rf = ModelArray(RandomForest(test_size=0.3, max_depth=1, n_estimators=1, random_state=42))
    rf.fit(df_new)

    pdp = ExplainerArray(rf, N=10, explanation_type=request.param[1])
    pdp.fit()
    t = pdp.transform()
    return t


def test_curve(pdp_transformed):
    try:
        str(pdp_transformed.results[list(pdp_transformed.results.keys())[0]])
        repr(pdp_transformed.results[list(pdp_transformed.results.keys())[0]])
    except (Exception,):
        assert False


def test_wrong_get_item(pdp_transformed):
    assert pdp_transformed['non-existing'] is None


def test_explainer_result_and_explainer_result_array(pdp_transformed):
    t = pdp_transformed

    try:
        str(t)
        repr(t)
        str(t[0])
        repr(t[0])
    except (Exception,):
        assert False


def test_plot_result_array(pdp_transformed):
    t = pdp_transformed
    column = list(pdp_transformed.results.keys())[0]

    try:
        t.plot(variable=column)
    except (Exception,):
        assert False


def test_plot_result_array_add_plot(pdp_transformed):
    t = pdp_transformed
    column = list(pdp_transformed.results.keys())[0]

    try:
        t1 = deepcopy(t)
        t1.results[column].y += 0.1
        t1.name += '_xx'
        t.plot(variable=column, add_plot=[t1])
    except (Exception,):
        assert False


def test_compare_result_array(pdp_transformed):
    t = pdp_transformed
    column1 = list(pdp_transformed.results.keys())[0]
    column2 = list(pdp_transformed.results.keys())[1]

    try:
        t1 = deepcopy(t)
        t1.results[column1].y += 0.1
        t1.name += '_xx'

        t.compare([t1, t1])
        t.compare([t1, t1], variable=column1)
        t.compare([t1, t1], variable=[column1, column2])
        x = t.compare([t1, t1], return_raw_per_variable=True)
    except (Exception,):
        assert False

    assert len(x) > 1


def test_compare_result_array_2(model_array):
    cols = list(model_array.get_models()[0].get_test_dataset().data.columns)
    column1 = cols[0]
    column2 = cols[1]

    pdp = ExplainerArray(model_array, N=10)
    pdp.fit()
    t = pdp.transform(variables=[column1])

    with pytest.raises(Exception):
        t[0].plot(variable=column2)


@pytest.fixture(
    scope='module',
    params=[
        (DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2, pd.concat([df_1_x for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            DatasetArray([
                Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
            ])
        ]), 'ALE'),
        (DatasetArray([
            Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            Dataset(name_2, pd.concat([df_1_x for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
            DatasetArray([
                Dataset(name_1, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)])),
                Dataset(name_2, pd.concat([df_1 for _ in range(5)]), pd.concat([target_1 for _ in range(5)]))
            ])
        ]), 'PDP')
    ]
)
def model_array2(request):
    df = request.param[0]
    df.remove_nans()
    transformator = TransformerArray(
        TransformerFromIMBLEARN(RandomUnderSampler(sampling_strategy=1, random_state=42)))
    transformator.fit(df)
    df_new = transformator.transform(df)

    rf = ModelArray(RandomForest(test_size=0.3, max_depth=1, n_estimators=1, random_state=42))
    rf.fit(df_new)

    pdp = ExplainerArray(rf, N=10, explanation_type=request.param[1])
    pdp.fit()
    t = pdp.transform()

    return rf, t


def test_array_of_arrays(model_array2):
    rf, t = model_array2
    try:
        str(t[0])
        repr(t[0])
    except (Exception,):
        assert False


def test_array_of_arrays_plot(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    try:
        t[0].plot(variable=column)

        t1 = deepcopy(t)
        t1[0].results[column].y += 0.1
        t1[0].name += '_xx'

        t[0].plot(variable=column, add_plot=[t1[0]])
    except (Exception,):
        assert False


def test_explainer_result_plot(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    try:
        t[2].plot(variables=column)
    except (Exception,):
        assert False


def test_array_of_arrays_plot_regex(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    try:
        t[2].plot(variables=column, model_filter=f"^{name_1}")
    except (Exception,):
        assert False


def test_array_of_arrays_plot_summary_regex(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    try:
        t[2].plot_summary(variables=column, model_filters=[f"^{name_1}", f"^{name_2}"])
    except (Exception,):
        assert False


def test_array_of_arrays_compare(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    column1 = rf.get_models()[0].get_train_dataset().data.columns[1]
    try:
        t1 = deepcopy(t)
        t1[0].results[column].y += 0.1
        t1[0].name += '_xx'

        t[0].compare([t1[0], t1[0]])
        t[0].compare([t1[0], t1[0]], variable=column)
        t[0].compare([t1[0], t1[0]], variable=[column, column1])
    except (Exception,):
        assert False


def test_array_of_arrays_transform_columns_not_in_each(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    column1 = rf.get_models()[0].get_train_dataset().data.columns[1]
    try:
        t1 = deepcopy(t)
        t1[0].results[column].y += 0.1
        t1[0].name += '_xx'

        # there are also datasets, which don't contain given columns
        pdp = ExplainerArray(rf, N=10, explanation_type=t.explanation_type)
        pdp.fit()

        pdp.transform(variables=[column, column1])
    except (Exception,):
        assert False


def test_array_of_arrays_transform_columns_not_in_each_plot(model_array2):
    rf, t = model_array2
    column = rf.get_models()[0].get_train_dataset().data.columns[0]
    column1 = rf.get_models()[1].get_train_dataset().data.columns[0]
    try:
        t1 = deepcopy(t)
        t1[0].results[column].y += 0.1
        t1[0].name += '_xx'

        # there are also datasets, which don't contain given columns
        pdp = ExplainerArray(rf, N=10, explanation_type=t.explanation_type)
        pdp.fit()

        t_2 = pdp.transform(variables=[column, column1])

        t_2[0].plot(variable=column)
        t_2[1].plot(variable=column1)
        assert len(t_2[1].results.items()) == 1
        assert len(t_2[0].results.items()) == 1

        t[2][0].plot(variable=column)

        with pytest.raises(Exception):
            t_2[0].plot(variable=column1)

    except (Exception,):
        assert False


def test_explainer_result_array(model_array2):
    _, t = model_array2

    try:
        str(t)
        repr(t)
    except (Exception,):
        assert False


def test_explainer_result_array_plot_fail(model_array2):
    _, t = model_array2

    # there are different columns
    with pytest.raises(Exception):
        t.plot()


def test_explainer_result_array_plot(model_array2):
    _, t = model_array2

    try:
        t[2].plot()
    except (Exception,):
        assert False


def test_explainer_result_array_get(model_array2):
    rf, t = model_array2

    try:
        x = t[[0, 1]]
        y = t[list(t.results)[0].name]
        z = t[0]
    except (Exception,):
        assert False

    assert len(x) == 2
    assert isinstance(x, ModelProfileExplanationArray)
    assert isinstance(y, ModelProfileExplanation)
    assert isinstance(z, ModelProfileExplanation)





import pandas as pd
import pytest

from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import RandomOverSampler
from copy import deepcopy

from edgaro.balancing.transformer import TransformerFromIMBLEARN, RandomUnderSampler as RandomUnderSampler_EDAGR, \
    RandomOverSampler as RandomOverSampler_EDGAR, Transformer
from edgaro.base.base_transformer import BaseTransformer
from edgaro.data.dataset import Dataset, DatasetFromOpenML
from edgaro.data.dataset_array import DatasetArray
from edgaro.balancing.transformer_array import TransformerArray

from .resources.objects import *

ds_openml_1 = DatasetFromOpenML(task_id=task_id_2, apikey=APIKEY)
ds_openml_2 = deepcopy(ds_openml_1)
ds_openml_2.name += '__xx'


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1),
    Dataset(name_2, df_2, target_2),
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2),
                  DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])]),
    DatasetArray([DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)], name='x'),
                  DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])])
])
def test_transformer_array(imblearn_sampler, ds):
    try:
        transformer = TransformerFromIMBLEARN(imblearn_sampler)
        array = TransformerArray(transformer)
        array.fit(ds)
        array.transform(ds)
        str(transformer)
        repr(transformer)
        str(array)
        repr(array)
    except (Exception,):
        assert False


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1),
    Dataset(name_2, df_2, target_2),
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)]),
])
def test_transform_keep_original(imblearn_sampler, ds):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, keep_original_dataset=True)
    array.fit(ds)
    transform_true = array.transform(ds)

    array.keep_original_dataset = False
    transform_false = array.transform(ds)

    assert len(transform_true) - 1 == len(transform_false)


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    Dataset(name_1, df_1, target_1),
    Dataset(name_2, df_2, target_2),
])
@pytest.mark.parametrize('sufix', [
    '_transformed',
    '_example_sufix'
])
def test_transformer_sufix(imblearn_sampler, ds, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)
    for dataset in out.datasets:
        assert dataset.name == ds.name + sufix


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])
])
@pytest.mark.parametrize('sufix', [
    '_transformed',
    '_example_sufix'
])
def test_transformer_sufix_2_datasetarray(imblearn_sampler, ds, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)

    s = sufix if isinstance(sufix, str) else '_transformed_array'
    assert out.name == ds.name + s

    expected_names = [dataset.name + sufix for dataset in ds.datasets]
    assert len(expected_names) == len(out)
    for k in range(len(out)):
        assert out[k].name == expected_names[k]


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_2, df_2, target_2)])
])
@pytest.mark.parametrize('sufix', [
    ['_transformed'],
    ['_example_sufix']
])
def test_transformer_sufix_tab_datasetarray(imblearn_sampler, ds, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)

    expected_names = [ds.datasets[i].name + sufix[i] for i in range(len(ds.datasets))]
    assert len(expected_names) == len(out)

    for k in range(len(out)):
        assert out[k].name == expected_names[k]


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('sufix', [
    ['_transformed_0', '_transformed_1'],
    ['_example_sufix_0', '_example_sufix_1']
])
def test_transformer_sufix_tab_2_datasetarray(imblearn_sampler, ds, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)

    expected_names = [ds.datasets[i].name + sufix[i] for i in range(len(ds.datasets))]
    assert len(expected_names) == len(out)

    for k in range(len(out)):
        assert out[k].name == expected_names[k]


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomOverSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray(
        [Dataset(name_1, df_1, target_1), Dataset(name_2, df_1, target_1), Dataset(name_1 + '3', df_1, target_1)])
])
@pytest.mark.parametrize('sufix', [
    ['_transformed_0', '_transformed_1'],
    ['_example_sufix_0', '_example_sufix_1']
])
def test_transformer_sufix_tab_datasetarray_wrong_length(imblearn_sampler, ds, sufix):
    with pytest.raises(Exception):
        transformer = TransformerFromIMBLEARN(imblearn_sampler)
        array = TransformerArray(transformer, dataset_suffixes=sufix)
        array.fit(ds)
        array.transform(ds)


@pytest.mark.parametrize('imblearn_sampler,ratio', [
    (RandomUnderSampler(sampling_strategy=1, random_state=42), 1),
    (RandomOverSampler(sampling_strategy=1, random_state=42), 1)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])
])
def test_imbalance_ratio(imblearn_sampler, ratio, ds):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)
    array.fit(ds)
    out = array.transform(ds)

    for ds_array in out:
        assert ds_array.imbalance_ratio == round(ratio, 2)


@pytest.mark.parametrize('imblearn_sampler,ratio', [
    (RandomUnderSampler(sampling_strategy=1, random_state=42), 1),
    (RandomUnderSampler(sampling_strategy=0.95, random_state=42), 0.95),
    (RandomOverSampler(sampling_strategy=1, random_state=42), 1),
    (RandomOverSampler(sampling_strategy=0.95, random_state=42), 0.95)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([deepcopy(ds_openml_2)]),
    DatasetArray([deepcopy(ds_openml_1), deepcopy(ds_openml_2)])
])
def test_imbalance_ratio_2(imblearn_sampler, ratio, ds):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)
    array.fit(ds)
    out = array.transform(ds)

    for ds_array in out:
        assert round(1 / ds_array.imbalance_ratio, 2) == round(ratio, 2)


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomUnderSampler(sampling_strategy=0.95, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('param', [
    {
        'sampling_strategy': [1, 0.98]
    },
    {
        'random_state': [1, 2],
        'sampling_strategy': [1, 0.98]
    },
    {
        'random_state': [1, 2],
        'imbalance_ratio': [1 / 1, 1 / 0.98]
    },
    {
        'random_state': [1, 2],
        'IR': [1 / 1, 1 / 0.98]
    }
])
def test_set_get_params(imblearn_sampler, ds, param):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)
    array.fit(ds)
    array.set_params(**param)
    array.transform(ds)

    tmp = []
    for i in range(len(param[list(param.keys())[0]])):
        tmp_dict = {}
        for key in param:
            tmp_dict[key] = param[key][i]
        tmp.append(tmp_dict)

    assert array.get_params() == tmp

    for i in range(len(array.transformers)):
        tmp = {}
        for key in param:
            if key == 'imbalance_ratio' or key == 'IR':
                tmp['sampling_strategy'] = 1 / param[key][i]
            else:
                tmp[key] = param[key][i]

        expected_params = list(tmp.items())
        existing_params = list(array.transformers[i].get_params().items())
        assert np.alltrue([p in existing_params for p in expected_params])


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomUnderSampler(sampling_strategy=0.95, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('param', [
    {
        'sampling_strategy': [[1, 0.98], [1, 0.98]]
    },
    {
        'random_state': [[1, 2], [1, 2]],
        'sampling_strategy': [[1, 0.98], [1, 0.98]]
    }
])
def test_set_get_params_2(imblearn_sampler, ds, param):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)
    array.set_params(**param)
    array.fit(ds)
    array.transform(ds)

    def transform_dict_to_lst(dct):
        lengths = [len(val) for key_, val in dct.items()]
        tmp_ = []
        for i_ in range(lengths[0]):
            tmp_dict_ = {}
            for key_ in dct:
                tmp_dict_[key_] = dct[key_][i_]
            tmp_.append(tmp_dict_)
        return tmp_

    out1 = transform_dict_to_lst(param)
    assert array.get_params() == out1

    for i in range(len(array.transformers)):
        expected_params = transform_dict_to_lst(out1[i])
        existing_params = array.transformers[i].get_params()
        assert np.alltrue([expected_params[k] in existing_params for k in range(len(expected_params))])
        assert np.alltrue([existing_params[k] in expected_params for k in range(len(existing_params))])


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomUnderSampler(sampling_strategy=0.95, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('param', [
    [
        [
            {
                'sampling_strategy': 0.98
            },
            {
                'sampling_strategy': 1
            }
        ] for _ in range(2)
    ],
    [
        [
            {
                'sampling_strategy': 0.98,
                'random_state': 1
            },
            {
                'sampling_strategy': 1,
                'random_state': 2
            }
        ] for _ in range(2)
    ]
])
def test_params_in_arguments(imblearn_sampler, ds, param):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, parameters=param)
    array.fit(ds)
    array.transform(ds)

    assert array.get_params() == param

    for i in range(len(array.transformers)):
        for j in range(len(array.transformers[i])):
            expected_params = list(param[i][j].items())
            existing_params = list(array.transformers[i][j].get_params().items())
            assert np.alltrue([p in existing_params for p in expected_params])


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomUnderSampler(sampling_strategy=0.95, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('param', [
    [
        [
            {
                'sampling_strategy': 0.98
            },
            {
                'sampling_strategy': 1
            }
        ] for _ in range(2)
    ],
    [
        [
            {
                'sampling_strategy': 0.98,
                'random_state': 1
            },
            {
                'sampling_strategy': 1,
                'random_state': 2
            }
        ] for _ in range(2)
    ]
])
@pytest.mark.parametrize('sufix', [
    ['_transformed_0x', '_transformed_1x'],
    ['_example_sufix_0', '_example_sufix_1']
])
def test_params_in_arguments_and_sufix(imblearn_sampler, ds, param, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, parameters=param, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)

    # params
    assert array.get_params() == param

    for i in range(len(array.transformers)):
        for j in range(len(array.transformers[i])):
            expected_params = list(param[i][j].items())
            existing_params = list(array.transformers[i][j].get_params().items())
            assert np.alltrue([p in existing_params for p in expected_params])

    # sufix
    assert len(param) == len(out)
    assert len(sufix) == len(out)

    for j in range(len(out)):
        for k in range(len(out[j])):
            out_tmp = out[j][k]
            assert out_tmp.name == ds[j].name + sufix[j] + '_' + str(k)


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42),
    RandomUnderSampler(sampling_strategy=0.95, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_2, df_1, target_1), Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), deepcopy(ds_openml_2)])
])
@pytest.mark.parametrize('param', [
    [
        [
            {
                'sampling_strategy': 0.98
            },
            {
                'sampling_strategy': 1
            }
        ] for _ in range(2)
    ],
    [
        [
            {
                'sampling_strategy': 0.98,
                'random_state': 1
            },
            {
                'sampling_strategy': 1,
                'random_state': 2
            }
        ] for _ in range(2)
    ]
])
@pytest.mark.parametrize('sufix', [
    ['_transformed_0'],
    ['_example_sufix_0']
])
def test_params_in_arguments_and_sufix_2(imblearn_sampler, ds, param, sufix):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, parameters=param, dataset_suffixes=sufix)
    array.fit(ds)
    out = array.transform(ds)

    # params
    assert array.get_params() == param

    for i in range(len(array.transformers)):
        for j in range(len(array.transformers[i])):
            expected_params = list(param[i][j].items())
            existing_params = list(array.transformers[i][j].get_params().items())
            assert np.alltrue([p in existing_params for p in expected_params])

    # sufix
    assert len(param) == len(out)

    for j in range(len(out)):
        for k in range(len(out[j])):
            out_tmp = out[j][k]
            assert out_tmp.name == ds[j].name + sufix[0] + '_' + str(j) + '_' + str(k)


def test_over_under_sampler():
    ds = Dataset(name_2, df_1, target_1)

    try:
        transformer_1 = RandomOverSampler_EDGAR()
        transformer_1.fit(ds)
        transformer_1.transform(ds)

        transformer_2 = RandomUnderSampler_EDAGR()
        transformer_2.fit(ds)
        transformer_2.transform(ds)
    except (Exception,):
        assert False


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])
])
def test_was_fitted(imblearn_sampler, ds):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)
    assert not array.was_fitted
    array.fit(ds)
    assert array.was_fitted


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)]),
    DatasetArray([Dataset(name_1, df_1, target_1), Dataset(name_2, df_2, target_2)])
])
def test_output_verbose(imblearn_sampler, ds, capsys):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer, verbose=True)
    array.fit(ds)
    array.transform(ds)

    captured = capsys.readouterr()
    assert f'TransformerArray {array.__repr__()} was fitted with {ds.name}' in captured.out
    assert f'TransformerArray {array.__repr__()} transformed with {ds.name}' in captured.out


@pytest.mark.parametrize('imblearn_sampler', [
    RandomUnderSampler(sampling_strategy=1, random_state=42)
])
@pytest.mark.parametrize('ds', [
    DatasetArray([Dataset(name_1, df_1, target_1)])
])
def test_base_transformer_get_set(imblearn_sampler, ds):
    transformer = TransformerFromIMBLEARN(imblearn_sampler)
    array = TransformerArray(transformer)

    assert isinstance(array.base_transformer, Transformer)

    try:
        ar = deepcopy(array)
        ar.base_transformer = deepcopy(imblearn_sampler)
        ar.fit(ds)
    except (Exception,):
        assert False

    with pytest.raises(Exception):
        ar = deepcopy(array)
        tmp = deepcopy(imblearn_sampler)
        tmp.__class__ = BaseTransformer
        ar.base_transformer = tmp
        x = ar.base_transformer



from abc import ABC, abstractmethod
from typing import Dict, Any

from edgaro.data.dataset import Dataset


class BaseTransformer(ABC):
    """
    The abstract class to define transformations for a single Dataset.
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def fit(self, dataset: Dataset) -> None:
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            The object to fit BaseTransformer on.
        """
        pass

    @abstractmethod
    def transform(self, dataset: Dataset) -> Dataset:
        """
        Transform the object.

        Parameters
        ----------
        dataset : Dataset
            The object to be transformed.

        Returns
        -------
        Dataset
            The transformed object.
        """
        pass

    @abstractmethod
    def set_params(self, **params) -> None:
        """
        Set params for BaseTransformer.

        Parameters
        ----------
        params : dict
            The parameters to be set.
        """
        pass

    @abstractmethod
    def get_params(self) -> Dict[str, Any]:
        """
        Get parameters of BaseTransformer.

        Returns
        -------
        Dict[str, Any]
            The parameters.
        """
        pass

    @property
    @abstractmethod
    def was_fitted(self) -> bool:
        """
        The information whether the BaseTransformer was fitted.

        Returns
        -------
        bool
        """
        pass

    def __str__(self) -> str:
        return f"BaseTransformer {self.__class__.__name__}"

    def __repr__(self) -> str:
        return f"<BaseTransformer {self.__class__.__name__}>"



from __future__ import annotations

import copy
import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from abc import ABC, abstractmethod
from typing import Dict, Union, Tuple, List, Optional, Literal
from numpy import ndarray
from matplotlib.axes import Axes
from scipy.stats import wilcoxon


class Curve:
    """
    The class which represents the PDP/ALE curve for one variable.

    Parameters
    ----------
    x : np.ndarray
        Points on 0X axis.
    y : np.ndarray
        Points on 0Y axis.

    Attributes
    ----------
    x : np.ndarray
        Points on 0X axis.
    y : np.ndarray
        Points on 0Y axis.
    """

    def __init__(self, x: Optional[ndarray, pd.Series], y: Optional[ndarray, pd.Series]) -> None:
        self.x = x if not isinstance(x, pd.Series) else x.reset_index(drop=True)
        self.y = y if not isinstance(y, pd.Series) else y.reset_index(drop=True)

    def __str__(self) -> str:
        return f"Curve with {len(self.x)} points\nx:\n" + str(self.x) + "\ny:\n" + str(self.y)

    def __repr__(self) -> str:
        return f"<Curve with {len(self.x)} points>"


class Explanation(ABC):
    def __int__(self) -> None:
        pass

    @abstractmethod
    def plot(self) -> None:
        pass

    @abstractmethod
    def compare(self, other: List[Explanation]) -> List[Union[float, list]]:
        pass


class ModelProfileExplanation(Explanation):
    """
    The class which represent the PDP/ALE curves for all variables in one Model.

    Parameters
    ----------
    results : Dict[str, Curve]
        A dictionary of pairs (column name, Curve object), which represents curves for all variables in one Model.
    name : str
        The name of ModelProfileExplanation. It is best if it is a Model name.
    categorical_columns : list[str]
        List of categorical variables.
    explanation_type : {'PDP', 'ALE'}, default='PDP'
        A curve type.

    Attributes
    ----------
    results : Dict[str, Curve]
        A dictionary of pairs (column name, Curve object), which represents curves for all variables in one Model.
    name : str
        The name of ModelProfileExplanation. It is best if it is a Model name.
    categorical_columns : list[str]
        List of categorical variables.
    explanation_type : {'PDP', 'ALE'}
        A curve type.

    """

    def __init__(self, results: Dict[str, Curve], name: str, categorical_columns: List[str],
                 explanation_type: Literal['PDP', 'ALE'] = 'PDP') -> None:
        self.results = results
        self.name = name
        self.categorical_columns = categorical_columns
        self.explanation_type = explanation_type

    def __getitem__(self, key: str) -> Optional[Curve]:
        if key in self.results.keys():
            return self.results[key]
        else:
            return None

    def plot(self, variable: Optional[str] = None, figsize: Optional[Tuple[int, int]] = (8, 8),
             add_plot: Optional[List[ModelProfileExplanation]] = None, ax: Optional[Axes] = None,
             show_legend: bool = True, y_lim: Optional[Tuple[float, float]] = None, metric_precision: int = 2,
             centered: bool = False) -> None:
        """
        The function plots the PDP/ALE curve.

        Parameters
        ----------
        variable : str, optional, default=None
            Variable for which the plot should be generated. If None, the first column is plotted.
        figsize : tuple(int, int), optional, default=(8, 8)
            Size of a figure.
        add_plot : list[ModelProfileExplanation], optional, default=None
            List of other ModelProfileExplanation objects that also contain the `variable` and should be plotted.
        ax : matplotlib.axes.Axes, optional, default=None
            The parameter should be passed if the plot is to be created in a certain Axis. In that situation, `figsize`
            parameter is ignored.
        show_legend : bool, default=True
            The parameter indicates whether the legend should be plotted.
        y_lim : tuple(float, float), optional, default=None
            The limits of 0Y axis.
        metric_precision : int, default=5
            Number of digits to round the value of `metric value*10^5`.
        centered : bool, default = False
            If True, the plots will be centered to start at 0.

        """
        if figsize is None and ax is None:
            figsize = (8, 8)

        if variable is None:
            variable = list(self.results.keys())[0]

        if add_plot is None:
            ModelProfileExplanation.__plot_not_add(self.results, self.categorical_columns, self.explanation_type, self.name,
                                                   variable, ax, figsize, show_legend, y_lim, centered)
        else:
            curve_base = self.results[variable]
            if curve_base is None:
                raise Exception('Variable is not available!')

            curves_add = [pdp.results[variable] for pdp in add_plot]
            if None in curves_add:
                warnings.warn(f'There is not variable {variable} in one of the added plots!')

            if len(curves_add) == 0:
                warnings.warn(f'None of the added plots have variable called {variable}!')
                ModelProfileExplanation.__plot_not_add(self.results, self.categorical_columns, self.explanation_type,
                                                       self.name, variable, ax, figsize, show_legend, y_lim, centered)
            else:
                self.__plot_add(variable, ax, figsize, curve_base, curves_add, add_plot, show_legend,
                                y_lim, metric_precision, centered)
        plt.ylabel(self.explanation_type)

    @staticmethod
    def __plot_not_add(results, categorical_columns, explanation_type, name, variable, ax, figsize, show_legend,
                       y_lim, centered) -> None:
        curve = results[variable]
        if curve is None:
            raise Exception('Variable is not available!')

        if ax is not None:
            plt.sca(ax)
        elif figsize is not None:
            plt.subplots(figsize=figsize)

        x, y = ModelProfileExplanation.__get_x_y(curve, centered)

        if variable not in categorical_columns:
            plt.plot(x, y)
        else:
            plt.bar(x, y)

        if explanation_type == 'PDP':
            plt.title("PDP curve for variable: " + variable)
        elif explanation_type == 'ALE':
            plt.title("ALE curve for variable: " + variable)
        else:
            raise Exception('Wrong curve type!')

        plt.xlabel(variable)
        if show_legend:
            plt.legend([name])
        if y_lim is not None:
            plt.ylim(y_lim)

    @staticmethod
    def __get_x_y(curve, centered):
        x = curve.x
        y = copy.deepcopy(curve.y)
        if centered:
            y = y - y[0]
        return x, y

    @staticmethod
    def __plot_add_continuous(ax, figsize, curve_base, curves_add, centered) -> None:
        if ax is not None:
            plt.sca(ax)
        elif figsize is not None:
            plt.subplots(figsize=figsize)

        x, y = ModelProfileExplanation.__get_x_y(curve_base, centered)
        plt.plot(x, y)
        for curve in curves_add:
            x, y = ModelProfileExplanation.__get_x_y(curve, centered)
            plt.plot(x, y)

    @staticmethod
    def __plot_add_categorical(name, variable, ax, figsize, curve_base, curves_add, add_plot_names, title, centered) -> None:
        _, y = ModelProfileExplanation.__get_x_y(curve_base, centered)

        df = pd.DataFrame({
            'x': curves_add[0].x,
            name: y
        })

        for i in range(len(curves_add)):
            _, y = ModelProfileExplanation.__get_x_y(curves_add[i], centered)
            df[add_plot_names[i]] = y

        if ax is not None:
            df.plot(x='x', kind='bar', ax=ax, legend=False, xlabel=variable, title=title)
        else:
            df.plot(x='x', kind='bar', figsize=figsize, legend=False, xlabel=variable, title=title)

        plt.xticks(rotation=0)

    def __plot_add(self, variable, ax, figsize, curve_base, curves_add, add_plot, show_legend,
                   y_lim, metric_precision, centered) -> None:
        curves_add = [c for c in curves_add if c is not None]
        add_plot_names = [c.name for c in add_plot if c.results[variable] is not None]

        p = 1
        for ind in range(len(add_plot_names)):
            if add_plot_names[ind] == self.name:
                add_plot_names[ind] += ('_' + str(p))
                p += 1

        if variable not in self.categorical_columns:
            ModelProfileExplanation.__plot_add_continuous(ax, figsize, curve_base, curves_add, centered)
            plt.title(f"{self.explanation_type} curve for variable: " + variable)
        else:
            ModelProfileExplanation.__plot_add_categorical(self.name, variable, ax, figsize,
                                                           curve_base, curves_add, add_plot_names,
                                                           f"{self.explanation_type} curve for variable: " + variable,
                                                           centered)

        if show_legend:
            plt.legend([self.name] + add_plot_names)
        if y_lim is not None:
            plt.ylim(y_lim)
        plt.xlabel(variable)

        compare_results = self.compare(add_plot, variable=variable)
        if not np.alltrue([r is None for r in compare_results]):
            if ax is None:
                ax = plt.gca()
            y_min, y_max = ax.get_ylim()
            x_min, x_max = ax.get_xlim()

            text = ""
            if compare_results[0] is not None:
                if len(add_plot) == 1:
                    text += 'SDD[$10^{-3}$]=' + f'{round(compare_results[0] * 10 ** 3, metric_precision)}'
                else:
                    text += 'ASDD[$10^{-3}$]=' + f'{round(compare_results[0] * 10 ** 3, metric_precision)}'

            if text != "":
                ax.text(
                    x_min + 0.7 * (x_max - x_min),
                    y_min + 0.2 * (y_max - y_min),
                    text,
                    fontsize='medium'
                )

    @staticmethod
    def __retrieve_explainer_results(inp, explain_results_in):
        if isinstance(inp, ModelProfileExplanation):
            explain_results_in.append(inp)
        else:
            if isinstance(inp, list):
                for inp_i in inp:
                    ModelProfileExplanation.__retrieve_explainer_results(inp_i, explain_results_in)
            else:
                for inp_i in inp.results:
                    ModelProfileExplanation.__retrieve_explainer_results(inp_i, explain_results_in)

    def compare(self, other: List[ModelProfileExplanation], variable: Optional[Union[str, List[str]]] = None,
                return_raw_per_variable: bool = False) -> List[Union[float, list]]:
        """
        The function calculates the metric to compare the curves for a given variable(s).

        Currently, there is only one comparison metric called `SDD` (Standard Deviation of Distances). It is the variance of
        the distances between curve in this object and curves in `other` in intermediate points.
        If there is more than one `other` object and `return_raw_per_variable=False`, the mean variance
        is returned (`ASDD` - Averaged SDD).

        Parameters
        ----------
        other : list[ModelProfileExplanation]
            List of ModelProfileExplanation objects to compare the curve against.
        variable : str, list[str], optional, default=None
            List of variable names to calculate the metric distances. If None, the metrics are calculated for
            all the columns in this object.
        return_raw_per_variable : bool, default=False
            If True, raw values for each variable are returned.

        Returns
        ----------
        list[float, list]

        """

        result_tab = []
        if isinstance(variable, str):
            if self[variable] is None:
                raise Exception('Variable is not available!')

            explain_results = []
            ModelProfileExplanation.__retrieve_explainer_results(other, explain_results)

            if np.all([o[variable] is None for o in explain_results]):
                raise Exception('Variable in \'other\' is not available!')

            result_tab = self.__calculate_metrics_one_variable(variable, explain_results)
        else:
            var = list(self.results.keys()) if variable is None else variable
            outs = [self.compare(variable=v, other=other) for v in var]

            if not return_raw_per_variable:
                for i in range(len(outs[0])):
                    result_tab.append(float(np.mean([
                        o[i] for o in outs
                    ])))
            else:
                result_tab = [tmp[0] for tmp in outs]

        return result_tab

    def __calculate_metrics_one_variable(self, variable: str, explain_results: List[ModelProfileExplanation]):

        def dist(res):
            y1 = res[variable].y
            y1 = y1.reset_index(drop=True) if isinstance(y1, pd.Series) else y1

            y2 = self[variable].y
            y2 = y2.reset_index(drop=True) if isinstance(y2, pd.Series) else y2
            return y1 - y2

        result_tab = []
        distances_to_original = [
            dist(res)
            for res in explain_results
        ]

        # fligner test
        # if len(explain_results) < 2:
        #    result_tab.append(None)
        # else:
        #    _, out = fligner(*distances_to_original)
        #    result_tab.append(out)

        # variance
        variances = [np.std(deltas) for deltas in distances_to_original]
        result_tab.append(np.mean(variances))

        # mean of abs differences
        # abs_deltas = [np.mean(np.abs(deltas)) for deltas in distances_to_original]
        # result_tab.append(np.mean(abs_deltas))

        return result_tab

    def __str__(self) -> str:
        return f"ModelProfileExplanation {self.name} for {len(self.results.keys())} variables: {list(self.results.keys())} with {self.explanation_type} curve type"

    def __repr__(self) -> str:
        return f"<ModelProfileExplanation {self.name} with {self.explanation_type} curve type>"


class ModelPartsExplanation(Explanation):
    """
    The class represents the Variable Importance for all variables in one Model.

    Parameters
    ----------
    results : Dict[str, float]
        A dictionary of pairs (column name, value), which represents Variable Importance for all variables in one Model.
    name : str
        The name of ModelProfileExplanation. It is best if it is a Model name.
    explanation_type : {'VI'}, default='VI'
        An explanation type.

    Attributes
    ----------
    results : Dict[str, float]
        A dictionary of pairs (column name, value), which represents Variable Importance for all variables in one Model.
    name : str
        The name of ModelProfileExplanation. It is best if it is a Model name.
    explanation_type : {'VI'}, default='VI'
        An explanation type.
    """

    def __init__(self, results: Dict[str, float], name: str, explanation_type: Literal['VI'] = 'VI') -> None:
        self.results = results
        self.name = name
        self.explanation_type = explanation_type

    def __getitem__(self, key: str) -> Optional[float]:
        if key in self.results.keys():
            return self.results[key]
        else:
            return None

    @staticmethod
    def __extract_res(obj, variable, max_variables):
        res = pd.DataFrame.from_dict(obj.results, orient='index').reset_index()
        res.columns = ['colname', 'val']
        res['colname'] = res['colname'].astype("string")
        res = res.sort_values('val')

        if variable is not None:
            res = res[res['colname'].isin(variable)]
            if res.shape[0] != len(variable):
                raise Exception('Wrong variable names were provided!')

        if max_variables is not None:
            r = min(max_variables, len(variable)) if variable is not None else max_variables
            r = min(r, res.shape[0])
            res = res.iloc[list(range(r))]

        return res

    def plot(self, variable: Optional[Union[str, List[str]]] = None,
             add_plot: Optional[List[ModelPartsExplanation]] = None, max_variables: Optional[int] = None,
             figsize: Optional[Tuple[int, int]] = (8, 8), ax: Optional[Axes] = None, show_legend: bool = False,
             x_lim: Optional[Tuple[float, float]] = None, metric_precision: int = 3) -> None:
        """
        The function plots the Variable Importance profile.

        Parameters
        ----------
        variable : str, list[str], optional, default=None
            Variable for which the VI should be plotted. If None, the all columns is plotted.
        figsize : tuple(int, int), optional, default=(8, 8)
            Size of a figure.
        add_plot : list[ModelPartsExplanation], optional, default=None
            List of other ModelPartsExplanation objects that also contain the `variable` and should be plotted.
        max_variables : int, optional, default=None
            Maximal number of variables from the current object to be taken into account.
        ax : matplotlib.axes.Axes, optional, default=None
            The parameter should be passed if the plot is to be created in a certain Axis. In that situation, `figsize`
            parameter is ignored.
        show_legend : bool, default=True
            The parameter indicates whether the legend should be plotted.
        x_lim : tuple(float, float), optional, default=None
            The limits of 0X axis.
        metric_precision : int, default=5
            Number of digits to round the value of the metric value.

        """

        res = ModelPartsExplanation.__extract_res(self, variable, max_variables)
        column_order = res['colname'].to_list()

        if isinstance(variable, str):
            variable = [variable]

        if ax is not None:
            plt.sca(ax)
        elif figsize is not None:
            plt.subplots(figsize=figsize)

        if add_plot is not None:
            def extraction(a):
                res_other = ModelPartsExplanation.__extract_res(a, variable, None)
                res_other['colname'] = res_other['colname'].astype("string")
                res_other = res_other.set_index('colname')
                res_other = res_other.loc[column_order]
                res_other = res_other.reset_index()
                return res_other

            group_height = 0.8
            y = res['colname'].astype("string")
            bar_height = group_height / (len(add_plot) + 1)
            y_axis = np.arange(len(y))

            plt.barh(y_axis, res['val'], bar_height, label=self.name)

            for i in range(len(add_plot)):
                p = extraction(add_plot[i])
                plt.barh(y_axis + (i + 1) * bar_height, p['val'], bar_height, label=add_plot[i].name)

            plt.yticks(y_axis + group_height / 2, y)
        else:
            plt.barh(res['colname'], res['val'])

        plt.title(f"Variable importance for {self.name}")

        plt.xlabel("Variable Importance")
        plt.ylabel("Variable")
        if show_legend:
            if add_plot is None:
                plt.legend([self.name])
            else:
                plt.legend()
        if x_lim is not None:
            plt.xlim(x_lim)

        if add_plot is not None:
            compare_results = self.compare(add_plot, variable=variable, max_variables=max_variables, return_raw=False)
            if not np.alltrue([r is None for r in compare_results]):
                if ax is None:
                    ax = plt.gca()
                y_min, y_max = ax.get_ylim()
                x_min, x_max = ax.get_xlim()

                text = ""
                if compare_results[0] is not None:
                    if len(add_plot) == 1:
                        text += 'p-value=' + f'{round(compare_results[0], metric_precision)}'
                    else:
                        text += 'MEAN(p-value)=' + f'{round(compare_results[0], metric_precision)}'

                if text != "":
                    ax.text(
                        x_min + 0.7 * (x_max - x_min),
                        y_min + 0.2 * (y_max - y_min),
                        text,
                        fontsize='medium'
                    )

    def compare(self, other: List[ModelPartsExplanation], variable: Optional[Union[str, List[str]]] = None,
                max_variables: Optional[int] = None, return_raw: bool = True) -> List[Union[float, list]]:
        """
        The function calculates the metric to compare model parts of two or more models.

        Currently, there is only one metric based on the Wilcoxon statistical test [3]_. The metric value is the
        p-value of this test, where the inputs are variable importance values.
        The idea is based on the approach presented in the article [4]_.

        Parameters
        ----------
        other : list[ModelPartsExplanation]
            List of ModelPartsExplanation objects to compare the curve against.
        variable : str, list[str], optional, default=None
            List of variable names to calculate the metric distances. If None, the metrics are calculated for
            all the columns in this object.
        max_variables : int, optional, default=None
            Maximal number of variables from the current object to be taken into account.
        return_raw : bool, default=True
            If True, the p-values are returned for each model. Otherwise, the mean value is returned.

        Returns
        -------
        list[float, list]

        References
        ----------
        .. [3] https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wilcoxon.html
        .. [4] https://www.sciencedirect.com/science/article/pii/S2666764922000145

        """

        out = []

        res = ModelPartsExplanation.__extract_res(self, variable, max_variables)

        column_order = res['colname'].to_list()
        for obj in other:
            res_other = pd.DataFrame.from_dict(obj.results, orient='index').reset_index()
            res_other.columns = ['colname', 'val']
            res_other['colname'] = res_other['colname'].astype("string")
            res_other = res_other.set_index('colname')
            res_other = res_other.loc[column_order]
            res_other = res_other.reset_index()

            stat_test = wilcoxon(res['val'].tolist(), res_other['val'].to_list())
            p_value = stat_test[1]
            out.append(p_value)

        if not return_raw:
            return [float(np.mean(out))]
        else:
            return out

    def __str__(self) -> str:
        return f"ModelPartsExplanation {self.name} for {len(self.results.keys())} variables: {list(self.results.keys())} with {self.explanation_type} curve type"

    def __repr__(self) -> str:
        return f"<ModelPartsExplanation {self.name} with {self.explanation_type} curve type>"
